
# -*- coding: utf-8 -*-

################################################################################
#                                                                              #
#   This file is part of the Bibolamazi Project.                               #
#   Copyright (C) 2014 by Philippe Faist                                       #
#   philippe.faist@bluewin.ch                                                  #
#                                                                              #
#   Bibolamazi is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU General Public License as published by       #
#   the Free Software Foundation, either version 3 of the License, or          #
#   (at your option) any later version.                                        #
#                                                                              #
#   Bibolamazi is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#   GNU General Public License for more details.                               #
#                                                                              #
#   You should have received a copy of the GNU General Public License          #
#   along with Bibolamazi.  If not, see <http://www.gnu.org/licenses/>.        #
#                                                                              #
################################################################################

# Py2/Py3 support
from __future__ import unicode_literals, print_function
from past.builtins import basestring
from future.utils import python_2_unicode_compatible, iteritems
from builtins import range
from builtins import str as unicodestr


from collections import namedtuple
import logging

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

import bibolamazi.init
from bibolamazi.core.helppages import htmlescape, forcewrap_long_lines

from .overlistbuttonwidget import OverListButtonWidgetBase
from .qtauto.ui_favoritesoverbtns import Ui_FavoritesOverBtns

logger = logging.getLogger(__name__)


class FavoriteCmd(object):
    def __init__(self, name, cmd, builtin=False):
        self.name = name
        self.cmd = cmd
        self.builtin = builtin


builtin_cmds = [
    FavoriteCmd(name="fixes (simple)",
                cmd=("""\
filter: fixes -dFixSpaceAfterEscape -dRemoveTypeFromPhd -dRemoveFileField -dEncodeUtf8ToLatex"""),
                builtin=True),
    FavoriteCmd(name="fixes (extensive)",
                cmd=("""\
filter: fixes -dFixSpaceAfterEscape
              -dEncodeUtf8ToLatex
              -dRemoveTypeFromPhd
              -sRemoveFullBraces=title
              -sRemoveFullBracesNotLang=german
              -sProtectNames="Maxwell,Landauer,Neumann,Szilard,Bell,I,II,IV,XIV,C++,Gram,ALPS,Ehrenfest,Pauli,Monte,Carlo,CVX,Finetti,Rényi,R{\\'e}nyi,R{\\'{e}}nyi,Golden-Thompson,QBism,Boltzmann,Birkhoff,Gibbs,Einstein,Podolsky,Rosen,Hamiltonian,Petz,AdS,CFT"
              -dRemoveFileField
              -sRenameLanguage=en:english,French:frenchb,German:german,Deutsch:german
              -dFixMendeleyBugUrls"""),
                builtin=True),
    FavoriteCmd(name="arxiv (for notes & verbose)",
                cmd="filter: arxiv -sMode=eprint -sUnpublishedMode=eprint -dThesesCountAsPublished -dWarnJournalRef",
                builtin=True),
    FavoriteCmd(name="arxiv (for articles & finalizing)",
                cmd="filter: arxiv -sMode=strip -sUnpublishedMode=unpublished-note",
                builtin=True),
    FavoriteCmd(name="url (standard cleanup)",
                cmd="filter: url -dStripAllIfDoiOrArxiv -dStripDoiUrl -dStripArxivUrl -dKeepFirstUrlOnly",
                builtin=True),
    FavoriteCmd(name="orderentries (alphabetically by cite key)",
                cmd="filter: orderentries -sOrder=alphabetical",
                builtin=True),
    ];

class FavoriteCmdsList(QObject):
    def __init__(self, parent):
        super(FavoriteCmdsList,self).__init__(parent)
        self.favlist = [];

    favChanged = pyqtSignal()

    def loadDefaults(self):
        self.favlist[:] = builtin_cmds

    def loadFromSettings(self, settings):
        settings.beginGroup("Favorites")

        self.favlist[:] = builtin_cmds
        
        siz = settings.beginReadArray("custom_favorite_cmds")
        for i in range(siz):
            settings.setArrayIndex(i)
            self.favlist.append(FavoriteCmd(name=self._get_settings_str(settings, "name"),
                                            cmd=self._get_settings_str(settings, "cmd")))
        settings.endArray()

        settings.endGroup()

        self.favChanged.emit()

    def _get_settings_str(self, settings, field):
        val = settings.value(field)
        if val is None:
            return ''
        return str(val)

    def saveToSettings(self, settings):
        settings.beginGroup("Favorites")

        custom_favlist = [x for x in self.favlist if not x.builtin]

        settings.beginWriteArray("custom_favorite_cmds", len(custom_favlist))
        for i in range(len(custom_favlist)):
            fav = custom_favlist[i]
            settings.setArrayIndex(i)
            settings.setValue("name", str(fav.name))
            settings.setValue("cmd", str(fav.cmd))
        settings.endArray()

        settings.endGroup()

    def addFavorite(self, favcmd):
        self.favlist += [favcmd]
        self.favChanged.emit()

    def moveFavorite(self, oldind, newind):
        if oldind == newind:
            return

        if oldind < newind:
            self.favlist[oldind:newind] = self.favlist[oldind+1:newind] + [self.favlist[oldind]]
        else: # oldind > newind
            self.favlist[newind:oldind+1] = [self.favlist[oldind]] + self.favlist[newind:oldind]

        self.favChanged.emit()




ROLE_CMD            = Qt.UserRole + 2014
ROLE_FAV_EDITABLE   = ROLE_CMD + 1
ROLE_FAVCMD_OBJECT  = ROLE_CMD + 2
ROLE_CMD_IS_BUILTIN = ROLE_CMD + 3

class FavoritesModel(QAbstractTableModel):
    def __init__(self, favcmds, parent):
        super(FavoritesModel, self).__init__(parent=parent)
        self._favcmds = favcmds;
        self._favcmds.favChanged.connect(self.modelReset)
        self._edit_mode = False

    def setEditMode(self, editmodeon):
        self._edit_mode = editmodeon;
        self.modelReset.emit()

    def rowCount(self, index=QModelIndex()):
        if index.isValid():
            return 0;
        return len(self._favcmds.favlist);

    def columnCount(self, index=QModelIndex()):
        if index.isValid():
            return 0;
        return 1;


    def flags(self, index):
        if self._edit_mode:
            if not index.isValid():
                return Qt.ItemIsDropEnabled
            flags = (Qt.ItemIsSelectable | Qt.ItemIsEnabled | Qt.ItemIsDragEnabled)
            row = index.row()
            if row >= 0 and row < len(self._favcmds.favlist):
                if not self._favcmds.favlist[row].builtin:
                    flags = (flags | Qt.ItemIsEditable)
            return flags
        return (Qt.ItemIsSelectable | Qt.ItemIsEnabled)


    def data(self, index, role=Qt.DisplayRole):

        if not index.isValid():
            return None
        
        col = index.column()
        row = index.row()

        if (row < 0 or row >= len(self._favcmds.favlist)):
            return None

        if role == Qt.BackgroundRole:
            if self._favcmds.favlist[row].builtin:
                return QBrush(QColor(180,180,120,60))
            else:
                if self._edit_mode:
                    return QBrush(QColor(255,230,230, 80))
                else:
                    return QBrush(QColor(235,255,255, 80))

            return None

        if (role == ROLE_FAV_EDITABLE):
            return self._edit_mode

        if (role == ROLE_CMD):
            return self._favcmds.favlist[row].cmd

        if (role == ROLE_CMD_IS_BUILTIN):
            return self._favcmds.favlist[row].builtin

        if (role == ROLE_FAVCMD_OBJECT):
            return self._favcmds.favlist[row]

        # argument name
        if (role == Qt.DisplayRole):
            return self._favcmds.favlist[row].name

        # editable name
        if (role == Qt.EditRole):
            return self._favcmds.favlist[row].name
        
        # tool-tip documentation
        if (role == Qt.ToolTipRole):
            return forcewrap_long_lines(self._favcmds.favlist[row].cmd)

        return None

    def setData(self, index, value, role=Qt.EditRole):
        if not index.isValid():
            return False
        
        col = index.column()
        row = index.row()

        if (row < 0 or row >= len(self._favcmds.favlist)):
            return False

        self._favcmds.favlist[row].name = str('' if value is None else value)
        self.modelReset.emit()
        return True
        

    def removeFavorite(self, row):
        if (isinstance(row, QModelIndex)):
            row = row.row()
        if row < 0 or row >= len(self._favcmds.favlist):
            return
        del self._favcmds.favlist[row]
        self.modelReset.emit()

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if (orientation == Qt.Vertical):
            return None

        if (section == 0):
            if (role == Qt.DisplayRole):
                return u""
            return None

        if (section == 1):
            if (role == Qt.DisplayRole):
                return u""
            return None

        return None

    def supportedDropActions(self):
        return Qt.MoveAction

    def mimeTypes(self):
        return [ "application/x-bibolamazi-internalmove-favorites" ];

    def mimeData(self, indexes):
        logger.longdebug("indexes=%r", indexes)
        if len(indexes) != 1:
            return None

        mimeData = QMimeData()
        mimeData.setData('application/x-bibolamazi-internalmove-favorites',
                         QByteArray(str('%d' %(indexes[0].row()))))
        logger.longdebug("mimeData=%r", mimeData)
        return mimeData

    def dropMimeData(self, data, action, row, column, parent):
        logger.longdebug(
            "Drop! data=%r (formats=%r), action=%r, row=%r, column=%r, parent=%r (%d,%d)",
            data, list(data.formats()), action, row, column, parent, parent.row(), parent.column()
        )

        if not data.hasFormat('application/x-bibolamazi-internalmove-favorites'):
            return False

        if row < 0 or parent.isValid():
            return False

        oldrow = int(data.data("application/x-bibolamazi-internalmove-favorites"))
        logger.longdebug("oldrow: %r", oldrow)

        self._favcmds.moveFavorite(oldrow, row);
        return True



class FavoritesItemDelegate(QStyledItemDelegate):

    def __init__(self, parent):
        super(FavoritesItemDelegate, self).__init__(parent=parent)

    def paint(self, painter, option, index):

        painter.save()

        options = QStyleOptionViewItem(option)
        self.initStyleOption(options, index)
        options.text = ""

        htmlitem = self._htmlitem(index)

        doc = QTextDocument()
        doc.setHtml(htmlitem)

        style = QApplication.style() if options.widget is None \
            else options.widget.style()
        style.drawControl(QStyle.CE_ItemViewItem, options, painter)

        ctx = QAbstractTextDocumentLayout.PaintContext()

        if option.state & QStyle.State_Selected:
            ctx.palette.setColor(QPalette.Text, option.palette.color(
                                 QPalette.Active, QPalette.HighlightedText))

        textRect = style.subElementRect(QStyle.SE_ItemViewItemText, options)
        painter.translate(textRect.topLeft())
        painter.setClipRect(textRect.translated(-textRect.topLeft()))
        doc.documentLayout().draw(painter, ctx)

        painter.restore()        

    def sizeHint(self, option, index):

        options = QStyleOptionViewItem(option)
        self.initStyleOption(options, index)
        options.text = ""

        htmlitem = self._htmlitem(index, bare=True)

        doc = QTextDocument()
        doc.setHtml(htmlitem)
        doc.setTextWidth(options.rect.width())
        return QSize(doc.idealWidth(), doc.size().height())

    def _htmlitem(self, index, bare=False):

        favcmd = index.data(ROLE_FAVCMD_OBJECT)

        html = (
            "<html><head/><body>"+
            "<p style=\"font-weight:bold; margin: 3px 0px 0px 0px;\">" + htmlescape(favcmd.name) + "</p>"
            )

        html += "<p style=\"font-weight:italic; font-size: 0.9em; color: #808080; margin: 5px 0px 5px 0px;\">"
        if not bare:
            html += htmlescape(favcmd.cmd.strip())
        else:
            html += "&nbsp;"
        html += "</p>"

        html += "</body></html>"

        return html


class FavoritesOverBtns(OverListButtonWidgetBase):
    def __init__(self, itemview):
        super(FavoritesOverBtns, self).__init__(itemview=itemview)

        self.ui = Ui_FavoritesOverBtns()
        self.ui.setupUi(self)
        

    insertCommand = pyqtSignal('QString')

    def show_status(self, idx):
        if not idx.isValid():
            return False

        if idx.data(ROLE_FAV_EDITABLE):
            # edit mode
            self.ui.btnInsert.hide()
            self.ui.btnEndEdit.show()
            self.ui.btnEdit.setVisible(True if not idx.data(ROLE_CMD_IS_BUILTIN) else False)
            self.ui.btnDelete.setVisible(True if not idx.data(ROLE_CMD_IS_BUILTIN) else False)
            return True

        # else: normal, not editable
        self.ui.btnInsert.show()
        self.ui.btnEdit.show()
        self.ui.btnEndEdit.hide()
        self.ui.btnDelete.hide()
        return True

    @pyqtSlot()
    def on_btnInsert_clicked(self):
        curidx = self.curIndex()
        if not curidx.isValid():
            return

        self.insertCommand.emit( "\n" + curidx.data(ROLE_CMD) + "\n" )

    @pyqtSlot()
    def on_btnEdit_clicked(self):
        curidx = self.curIndex()
        if curidx.isValid() and curidx.data(ROLE_FAV_EDITABLE):
            # already in edit mode, initiate edit of item text:
            self.itemView().edit(curidx)
            return

        self.itemView().model().setEditMode(True)
        self.itemView().setCurrentIndex(self.curIndex())
        self.updateDisplay()

    @pyqtSlot()
    def on_btnEndEdit_clicked(self):
        self.itemView().model().setEditMode(False)
        self.updateDisplay()


    @pyqtSlot()
    def on_btnDelete_clicked(self):
        curidx = self.curIndex()
        if curidx.data(ROLE_CMD_IS_BUILTIN):
            QMessageBox.critical(self, "Built-in", "You can't remove a built-in favorite command")
        if QMessageBox.question(self, "Confirm deletion",
                                "Are you sure you want to delete this favorite command?",
                                QMessageBox.Yes|QMessageBox.No,
                                QMessageBox.No) != QMessageBox.Yes:
            return

        curidx.model().removeFavorite(curidx)
        self.updateDisplay()


    def get_widget_rect(self, rect):
        return rect



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'settingswidget.ui'
#
# Created by: PyQt5 UI code generator 5.12.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_SettingsWidget(object):
    def setupUi(self, SettingsWidget):
        SettingsWidget.setObjectName("SettingsWidget")
        SettingsWidget.resize(431, 513)
        self.gridLayout_4 = QtWidgets.QGridLayout(SettingsWidget)
        self.gridLayout_4.setContentsMargins(-1, 0, -1, -1)
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setContentsMargins(-1, 10, -1, -1)
        self.horizontalLayout_2.setSpacing(0)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem)
        self.btnTabGeneral = QtWidgets.QPushButton(SettingsWidget)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/pic/generalsettings.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.btnTabGeneral.setIcon(icon)
        self.btnTabGeneral.setIconSize(QtCore.QSize(64, 64))
        self.btnTabGeneral.setCheckable(True)
        self.btnTabGeneral.setAutoDefault(False)
        self.btnTabGeneral.setFlat(True)
        self.btnTabGeneral.setObjectName("btnTabGeneral")
        self.horizontalLayout_2.addWidget(self.btnTabGeneral)
        self.btnTabLocalRepos = QtWidgets.QPushButton(SettingsWidget)
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/pic/configlocalrepos.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.btnTabLocalRepos.setIcon(icon1)
        self.btnTabLocalRepos.setIconSize(QtCore.QSize(64, 64))
        self.btnTabLocalRepos.setCheckable(True)
        self.btnTabLocalRepos.setAutoDefault(False)
        self.btnTabLocalRepos.setFlat(True)
        self.btnTabLocalRepos.setObjectName("btnTabLocalRepos")
        self.horizontalLayout_2.addWidget(self.btnTabLocalRepos)
        self.btnTabRemoteRepos = QtWidgets.QPushButton(SettingsWidget)
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/pic/remoterepos.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.btnTabRemoteRepos.setIcon(icon2)
        self.btnTabRemoteRepos.setIconSize(QtCore.QSize(64, 64))
        self.btnTabRemoteRepos.setCheckable(True)
        self.btnTabRemoteRepos.setAutoDefault(False)
        self.btnTabRemoteRepos.setFlat(True)
        self.btnTabRemoteRepos.setObjectName("btnTabRemoteRepos")
        self.horizontalLayout_2.addWidget(self.btnTabRemoteRepos)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem1)
        self.gridLayout_4.addLayout(self.horizontalLayout_2, 0, 0, 1, 1)
        self.btns = QtWidgets.QDialogButtonBox(SettingsWidget)
        self.btns.setOrientation(QtCore.Qt.Horizontal)
        self.btns.setStandardButtons(QtWidgets.QDialogButtonBox.Ok)
        self.btns.setObjectName("btns")
        self.gridLayout_4.addWidget(self.btns, 3, 0, 1, 1)
        self.tabs = QtWidgets.QStackedWidget(SettingsWidget)
        self.tabs.setObjectName("tabs")
        self.pageGeneral = QtWidgets.QWidget()
        self.pageGeneral.setObjectName("pageGeneral")
        self.gridLayout_5 = QtWidgets.QGridLayout(self.pageGeneral)
        self.gridLayout_5.setContentsMargins(0, 0, 0, 0)
        self.gridLayout_5.setObjectName("gridLayout_5")
        self.wGeneral = QtWidgets.QGroupBox(self.pageGeneral)
        self.wGeneral.setObjectName("wGeneral")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.wGeneral)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.label = QtWidgets.QLabel(self.wGeneral)
        self.label.setObjectName("label")
        self.gridLayout_2.addWidget(self.label, 0, 0, 1, 1)
        self.chkHideStartupWindow = QtWidgets.QCheckBox(self.wGeneral)
        self.chkHideStartupWindow.setObjectName("chkHideStartupWindow")
        self.gridLayout_2.addWidget(self.chkHideStartupWindow, 1, 0, 1, 2)
        self.spnMaxRecentFiles = QtWidgets.QSpinBox(self.wGeneral)
        self.spnMaxRecentFiles.setMaximum(31)
        self.spnMaxRecentFiles.setObjectName("spnMaxRecentFiles")
        self.gridLayout_2.addWidget(self.spnMaxRecentFiles, 0, 1, 1, 1)
        self.gridLayout_5.addWidget(self.wGeneral, 0, 0, 1, 1)
        spacerItem2 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout_5.addItem(spacerItem2, 1, 0, 1, 1)
        self.tabs.addWidget(self.pageGeneral)
        self.pageLocalRepos = QtWidgets.QWidget()
        self.pageLocalRepos.setObjectName("pageLocalRepos")
        self.gridLayout_7 = QtWidgets.QGridLayout(self.pageLocalRepos)
        self.gridLayout_7.setContentsMargins(0, 0, 0, 0)
        self.gridLayout_7.setObjectName("gridLayout_7")
        self.groupBox_2 = QtWidgets.QGroupBox(self.pageLocalRepos)
        self.groupBox_2.setObjectName("groupBox_2")
        self.gridLayout = QtWidgets.QGridLayout(self.groupBox_2)
        self.gridLayout.setObjectName("gridLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.btnFilterPackageAdd = QtWidgets.QPushButton(self.groupBox_2)
        self.btnFilterPackageAdd.setObjectName("btnFilterPackageAdd")
        self.horizontalLayout.addWidget(self.btnFilterPackageAdd)
        self.btnFilterPackageRemove = QtWidgets.QPushButton(self.groupBox_2)
        self.btnFilterPackageRemove.setObjectName("btnFilterPackageRemove")
        self.horizontalLayout.addWidget(self.btnFilterPackageRemove)
        self.gridLayout.addLayout(self.horizontalLayout, 1, 0, 1, 2)
        self.lstFilterPackages = QtWidgets.QTreeView(self.groupBox_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lstFilterPackages.sizePolicy().hasHeightForWidth())
        self.lstFilterPackages.setSizePolicy(sizePolicy)
        self.lstFilterPackages.setIndentation(0)
        self.lstFilterPackages.setObjectName("lstFilterPackages")
        self.gridLayout.addWidget(self.lstFilterPackages, 0, 0, 1, 1)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.btnFilterPackageMoveUp = QtWidgets.QToolButton(self.groupBox_2)
        self.btnFilterPackageMoveUp.setObjectName("btnFilterPackageMoveUp")
        self.verticalLayout_3.addWidget(self.btnFilterPackageMoveUp)
        self.btnFilterPackageMoveDown = QtWidgets.QToolButton(self.groupBox_2)
        self.btnFilterPackageMoveDown.setObjectName("btnFilterPackageMoveDown")
        self.verticalLayout_3.addWidget(self.btnFilterPackageMoveDown)
        spacerItem3 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Preferred)
        self.verticalLayout_3.addItem(spacerItem3)
        self.gridLayout.addLayout(self.verticalLayout_3, 0, 1, 1, 1)
        self.gridLayout_7.addWidget(self.groupBox_2, 0, 0, 1, 1)
        spacerItem4 = QtWidgets.QSpacerItem(20, 10, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout_7.addItem(spacerItem4, 1, 0, 1, 1)
        self.tabs.addWidget(self.pageLocalRepos)
        self.pageRemoteRepos = QtWidgets.QWidget()
        self.pageRemoteRepos.setObjectName("pageRemoteRepos")
        self.gridLayout_6 = QtWidgets.QGridLayout(self.pageRemoteRepos)
        self.gridLayout_6.setContentsMargins(0, 0, 0, 0)
        self.gridLayout_6.setObjectName("gridLayout_6")
        self.groupBox_3 = QtWidgets.QGroupBox(self.pageRemoteRepos)
        self.groupBox_3.setObjectName("groupBox_3")
        self.gridLayout_8 = QtWidgets.QGridLayout(self.groupBox_3)
        self.gridLayout_8.setObjectName("gridLayout_8")
        self.chkRemoteAllow = QtWidgets.QCheckBox(self.groupBox_3)
        self.chkRemoteAllow.setObjectName("chkRemoteAllow")
        self.gridLayout_8.addWidget(self.chkRemoteAllow, 0, 0, 1, 1)
        self.label_2 = QtWidgets.QLabel(self.groupBox_3)
        self.label_2.setWordWrap(True)
        self.label_2.setObjectName("label_2")
        self.gridLayout_8.addWidget(self.label_2, 1, 0, 1, 1)
        self.gridLayout_6.addWidget(self.groupBox_3, 1, 0, 1, 1)
        spacerItem5 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout_6.addItem(spacerItem5, 3, 0, 1, 1)
        self.grpGithubAuth = QtWidgets.QGroupBox(self.pageRemoteRepos)
        self.grpGithubAuth.setObjectName("grpGithubAuth")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.grpGithubAuth)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.chkGithubAuth = QtWidgets.QCheckBox(self.grpGithubAuth)
        self.chkGithubAuth.setObjectName("chkGithubAuth")
        self.gridLayout_3.addWidget(self.chkGithubAuth, 0, 0, 1, 2)
        self.lblGithubAuthStatus = QtWidgets.QLabel(self.grpGithubAuth)
        self.lblGithubAuthStatus.setTextFormat(QtCore.Qt.RichText)
        self.lblGithubAuthStatus.setObjectName("lblGithubAuthStatus")
        self.gridLayout_3.addWidget(self.lblGithubAuthStatus, 1, 0, 1, 2)
        self.gridLayout_6.addWidget(self.grpGithubAuth, 2, 0, 1, 1)
        self.tabs.addWidget(self.pageRemoteRepos)
        self.gridLayout_4.addWidget(self.tabs, 1, 0, 1, 1)

        self.retranslateUi(SettingsWidget)
        self.tabs.setCurrentIndex(0)
        self.btns.accepted.connect(SettingsWidget.accept)
        self.btns.rejected.connect(SettingsWidget.reject)
        QtCore.QMetaObject.connectSlotsByName(SettingsWidget)

    def retranslateUi(self, SettingsWidget):
        _translate = QtCore.QCoreApplication.translate
        SettingsWidget.setWindowTitle(_translate("SettingsWidget", "Settings"))
        self.btnTabGeneral.setToolTip(_translate("SettingsWidget", "General settings"))
        self.btnTabLocalRepos.setToolTip(_translate("SettingsWidget", "Configure local filter packages"))
        self.btnTabRemoteRepos.setToolTip(_translate("SettingsWidget", "Remote filter packages"))
        self.wGeneral.setTitle(_translate("SettingsWidget", "General"))
        self.label.setText(_translate("SettingsWidget", "Remember recent files:"))
        self.chkHideStartupWindow.setText(_translate("SettingsWidget", "Hide startup window after opening a document"))
        self.groupBox_2.setTitle(_translate("SettingsWidget", "Configure local filter packages"))
        self.btnFilterPackageAdd.setText(_translate("SettingsWidget", "Add filter package ..."))
        self.btnFilterPackageRemove.setText(_translate("SettingsWidget", "Forget package"))
        self.btnFilterPackageMoveUp.setText(_translate("SettingsWidget", "▲"))
        self.btnFilterPackageMoveDown.setText(_translate("SettingsWidget", "▼"))
        self.groupBox_3.setTitle(_translate("SettingsWidget", "Remote filter packages"))
        self.chkRemoteAllow.setText(_translate("SettingsWidget", "Allow access to remote filter packages"))
        self.label_2.setText(_translate("SettingsWidget", "Warning: filters are python scripts that can execute arbitrary code. Only run filters from sources you trust."))
        self.grpGithubAuth.setTitle(_translate("SettingsWidget", "Github repositories as filter packages"))
        self.chkGithubAuth.setText(_translate("SettingsWidget", "Use authentication (e.g. access private repositories)"))
        self.lblGithubAuthStatus.setText(_translate("SettingsWidget", "&lt;auth status here...>"))


from . import bibolamazi_res_rc




# -*- coding: utf-8 -*-

################################################################################
#                                                                              #
#   This file is part of the Bibolamazi Project.                               #
#   Copyright (C) 2013 by Philippe Faist                                       #
#   philippe.faist@bluewin.ch                                                  #
#                                                                              #
#   Bibolamazi is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU General Public License as published by       #
#   the Free Software Foundation, either version 3 of the License, or          #
#   (at your option) any later version.                                        #
#                                                                              #
#   Bibolamazi is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#   GNU General Public License for more details.                               #
#                                                                              #
#   You should have received a copy of the GNU General Public License          #
#   along with Bibolamazi.  If not, see <http://www.gnu.org/licenses/>.        #
#                                                                              #
################################################################################

# Py2/Py3 support
from __future__ import unicode_literals, print_function
from past.builtins import basestring
from future.utils import python_2_unicode_compatible, iteritems
from builtins import range
from builtins import str as unicodestr


from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

import bibolamazi.init
from .qtauto.ui_overlistbuttonwidget import Ui_OverListButtonWidget


ROLE_OVERBUTTON = Qt.UserRole + 137

OVERBUTTON_NONE = 0
OVERBUTTON_ADD = 1
OVERBUTTON_REMOVE = 2

ROLE_ARGNAME = Qt.UserRole + 138


class OverListButtonWidgetBase(QWidget):
    def __init__(self, itemview):
        super(OverListButtonWidgetBase, self).__init__(itemview.viewport())
        self.hide()

        self._view = itemview
        self._view_viewport = self._view.viewport()
        self._view_viewport.setMouseTracking(True)
        self._view_viewport.installEventFilter(self)

        self._lastpos = None
        self._curidx = None

    def curIndex(self):
        return self._curidx

    def itemView(self):
        return self._view

    def eventFilter(self, obj, event):

        if (obj == self._view_viewport):
            if (event.type() == QEvent.FocusOut):
                self.hide()
            if (event.type() == QEvent.MouseMove):
                self.updateDisplay(event.pos())
            if (event.type() == QEvent.Leave):
                self.updateDisplay(False)

        return super(OverListButtonWidgetBase, self).eventFilter(obj, event)

    @pyqtSlot()
    def updateDisplay(self, pos=None):
        if (pos is None):
            pos = self._lastpos
        elif pos is False:
            pos = None
        self._lastpos = pos

        if pos is None:
            self._disappear()
            return
            
        idx = self._view.indexAt(pos)
        if (self.show_status(idx)):
            self._appear(idx)
            return True
        # else:
        self._disappear()
        return False


    def show_status(self, index):
        return False

    def _disappear(self):
        self._curidx = None
        self.hide()

    def _appear(self, idx):
        self.show()
        self.setGeometry(self.get_widget_rect(self._view.visualRect(idx)))
        self._curidx = idx

    def get_widget_rect(self, rect):
        rect2 = QRect(rect);
        sh = self.minimumSizeHint()
        rect2.setLeft(rect.right()-sh.width())
        return rect2





class OverListButtonWidget(OverListButtonWidgetBase):
    def __init__(self, itemview):
        super(OverListButtonWidget, self).__init__(itemview)

        self.ui = Ui_OverListButtonWidget()
        self.ui.setupUi(self)


    addClicked = pyqtSignal('QString')
    editClicked = pyqtSignal('QString')
    removeClicked = pyqtSignal('QString')
    addIndexClicked = pyqtSignal('QModelIndex')
    editIndexClicked = pyqtSignal('QModelIndex')
    removeIndexClicked = pyqtSignal('QModelIndex')
    

    def show_status(self, idx):
        if (not idx.isValid()):
            return False
        
        whichbtn = idx.data(ROLE_OVERBUTTON)
        if whichbtn is None:
            return False
        
        if (whichbtn == OVERBUTTON_ADD):
            self.ui.btnAdd.show()
            self.ui.btnEdit.hide()
            self.ui.btnRemove.hide()
            return True
        if (whichbtn == OVERBUTTON_REMOVE):
            self.ui.btnAdd.hide()
            self.ui.btnEdit.show()
            self.ui.btnRemove.show()
            return True

        return False
    

    @pyqtSlot()
    def on_btnAdd_clicked(self):
        curidx = self.curIndex()
        if (curidx is None):
            return
        self.addIndexClicked.emit(curidx)
        self.addClicked.emit(curidx.data(ROLE_ARGNAME))

    @pyqtSlot()
    def on_btnEdit_clicked(self):
        curidx = self.curIndex()
        if (curidx is None):
            return
        self.editIndexClicked.emit(curidx)
        self.editClicked.emit(curidx.data(ROLE_ARGNAME))

    @pyqtSlot()
    def on_btnRemove_clicked(self):
        curidx = self.curIndex()
        if (curidx is None):
            return
        self.removeIndexClicked.emit(curidx)
        self.removeClicked.emit(curidx.data(ROLE_ARGNAME))


    
        



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'helpbrowser.ui'
#
# Created by: PyQt5 UI code generator 5.7.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_HelpBrowser(object):
    def setupUi(self, HelpBrowser):
        HelpBrowser.setObjectName("HelpBrowser")
        HelpBrowser.resize(600, 600)
        self.lyt = QtWidgets.QGridLayout(HelpBrowser)
        self.lyt.setContentsMargins(0, 0, 0, 0)
        self.lyt.setObjectName("lyt")
        self.tabs = QtWidgets.QTabWidget(HelpBrowser)
        self.tabs.setDocumentMode(True)
        self.tabs.setTabsClosable(True)
        self.tabs.setMovable(True)
        self.tabs.setObjectName("tabs")
        self.lyt.addWidget(self.tabs, 0, 0, 1, 1)

        self.retranslateUi(HelpBrowser)
        self.tabs.setCurrentIndex(-1)
        QtCore.QMetaObject.connectSlotsByName(HelpBrowser)

    def retranslateUi(self, HelpBrowser):
        _translate = QtCore.QCoreApplication.translate
        HelpBrowser.setWindowTitle(_translate("HelpBrowser", "Bibolamazi Help & Reference"))




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'filterinstanceeditor.ui'
#
# Created by: PyQt5 UI code generator 5.12.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_FilterInstanceEditor(object):
    def setupUi(self, FilterInstanceEditor):
        FilterInstanceEditor.setObjectName("FilterInstanceEditor")
        FilterInstanceEditor.resize(403, 474)
        self.gridLayout = QtWidgets.QGridLayout(FilterInstanceEditor)
        self.gridLayout.setObjectName("gridLayout")
        self.lstOptions = QtWidgets.QTreeView(FilterInstanceEditor)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.lstOptions.sizePolicy().hasHeightForWidth())
        self.lstOptions.setSizePolicy(sizePolicy)
        self.lstOptions.setVerticalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        self.lstOptions.setHorizontalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        self.lstOptions.setIndentation(0)
        self.lstOptions.setItemsExpandable(False)
        self.lstOptions.setAllColumnsShowFocus(True)
        self.lstOptions.setObjectName("lstOptions")
        self.gridLayout.addWidget(self.lstOptions, 3, 0, 1, 2)
        self.lblOptionHelp = QtWidgets.QLabel(FilterInstanceEditor)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lblOptionHelp.sizePolicy().hasHeightForWidth())
        self.lblOptionHelp.setSizePolicy(sizePolicy)
        self.lblOptionHelp.setMinimumSize(QtCore.QSize(0, 40))
        self.lblOptionHelp.setText("")
        self.lblOptionHelp.setTextFormat(QtCore.Qt.RichText)
        self.lblOptionHelp.setWordWrap(True)
        self.lblOptionHelp.setObjectName("lblOptionHelp")
        self.gridLayout.addWidget(self.lblOptionHelp, 4, 0, 1, 2)
        self.lblErrorMsg = QtWidgets.QLabel(FilterInstanceEditor)
        self.lblErrorMsg.setStyleSheet("color: rgb(198, 0, 0)")
        self.lblErrorMsg.setTextFormat(QtCore.Qt.RichText)
        self.lblErrorMsg.setWordWrap(True)
        self.lblErrorMsg.setObjectName("lblErrorMsg")
        self.gridLayout.addWidget(self.lblErrorMsg, 1, 0, 1, 2)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.label = QtWidgets.QLabel(FilterInstanceEditor)
        self.label.setObjectName("label")
        self.horizontalLayout.addWidget(self.label)
        self.lblFilterName = QtWidgets.QLabel(FilterInstanceEditor)
        self.lblFilterName.setStyleSheet("QLabel { font-weight: bold }")
        self.lblFilterName.setText("")
        self.lblFilterName.setObjectName("lblFilterName")
        self.horizontalLayout.addWidget(self.lblFilterName)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.btnFilterHelp = QtWidgets.QPushButton(FilterInstanceEditor)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/pic/question.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.btnFilterHelp.setIcon(icon)
        self.btnFilterHelp.setObjectName("btnFilterHelp")
        self.horizontalLayout.addWidget(self.btnFilterHelp)
        self.btnAddFavorite = QtWidgets.QPushButton(FilterInstanceEditor)
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/pic/bookmark.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.btnAddFavorite.setIcon(icon1)
        self.btnAddFavorite.setObjectName("btnAddFavorite")
        self.horizontalLayout.addWidget(self.btnAddFavorite)
        self.gridLayout.addLayout(self.horizontalLayout, 2, 0, 1, 2)

        self.retranslateUi(FilterInstanceEditor)
        QtCore.QMetaObject.connectSlotsByName(FilterInstanceEditor)

    def retranslateUi(self, FilterInstanceEditor):
        _translate = QtCore.QCoreApplication.translate
        FilterInstanceEditor.setWindowTitle(_translate("FilterInstanceEditor", "Form"))
        self.lblErrorMsg.setText(_translate("FilterInstanceEditor", "Error: ..."))
        self.label.setText(_translate("FilterInstanceEditor", "Filter:"))
        self.btnFilterHelp.setText(_translate("FilterInstanceEditor", "info"))
        self.btnAddFavorite.setToolTip(_translate("FilterInstanceEditor", "Add this full command line to your favorites"))
        self.btnAddFavorite.setText(_translate("FilterInstanceEditor", "add"))


from . import bibolamazi_res_rc



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'filterpackagepatheditor.ui'
#
# Created by: PyQt5 UI code generator 5.12.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_FilterPackagePathEditor(object):
    def setupUi(self, FilterPackagePathEditor):
        FilterPackagePathEditor.setObjectName("FilterPackagePathEditor")
        FilterPackagePathEditor.resize(309, 390)
        self.verticalLayout = QtWidgets.QVBoxLayout(FilterPackagePathEditor)
        self.verticalLayout.setObjectName("verticalLayout")
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem)
        self.label = QtWidgets.QLabel(FilterPackagePathEditor)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.lblInfo = QtWidgets.QLabel(FilterPackagePathEditor)
        self.lblInfo.setMinimumSize(QtCore.QSize(0, 50))
        self.lblInfo.setStyleSheet("background-color: rgba(0, 0, 120, 25);")
        self.lblInfo.setTextFormat(QtCore.Qt.RichText)
        self.lblInfo.setWordWrap(True)
        self.lblInfo.setObjectName("lblInfo")
        self.verticalLayout.addWidget(self.lblInfo)
        self.btnSetLocalPackage = QtWidgets.QPushButton(FilterPackagePathEditor)
        self.btnSetLocalPackage.setObjectName("btnSetLocalPackage")
        self.verticalLayout.addWidget(self.btnSetLocalPackage)
        self.btnSetGithubRepo = QtWidgets.QPushButton(FilterPackagePathEditor)
        self.btnSetGithubRepo.setObjectName("btnSetGithubRepo")
        self.verticalLayout.addWidget(self.btnSetGithubRepo)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem1)

        self.retranslateUi(FilterPackagePathEditor)
        QtCore.QMetaObject.connectSlotsByName(FilterPackagePathEditor)

    def retranslateUi(self, FilterPackagePathEditor):
        _translate = QtCore.QCoreApplication.translate
        FilterPackagePathEditor.setWindowTitle(_translate("FilterPackagePathEditor", "Form"))
        self.label.setText(_translate("FilterPackagePathEditor", "Imported filter package"))
        self.lblInfo.setText(_translate("FilterPackagePathEditor", "path"))
        self.btnSetLocalPackage.setText(_translate("FilterPackagePathEditor", "set local python package ..."))
        self.btnSetGithubRepo.setText(_translate("FilterPackagePathEditor", "set github repository ..."))





# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'githubauthenticationdialog.ui'
#
# Created by: PyQt5 UI code generator 5.12.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_GithubAuthenticationDialog(object):
    def setupUi(self, GithubAuthenticationDialog):
        GithubAuthenticationDialog.setObjectName("GithubAuthenticationDialog")
        GithubAuthenticationDialog.resize(511, 535)
        self.verticalLayout = QtWidgets.QVBoxLayout(GithubAuthenticationDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        spacerItem = QtWidgets.QSpacerItem(20, 189, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem)
        self.lbl = QtWidgets.QLabel(GithubAuthenticationDialog)
        self.lbl.setTextFormat(QtCore.Qt.RichText)
        self.lbl.setWordWrap(True)
        self.lbl.setOpenExternalLinks(True)
        self.lbl.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByMouse)
        self.lbl.setObjectName("lbl")
        self.verticalLayout.addWidget(self.lbl)
        spacerItem1 = QtWidgets.QSpacerItem(20, 10, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self.verticalLayout.addItem(spacerItem1)
        self.txtToken = QtWidgets.QLineEdit(GithubAuthenticationDialog)
        self.txtToken.setObjectName("txtToken")
        self.verticalLayout.addWidget(self.txtToken)
        spacerItem2 = QtWidgets.QSpacerItem(20, 188, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem2)
        self.btns = QtWidgets.QDialogButtonBox(GithubAuthenticationDialog)
        self.btns.setOrientation(QtCore.Qt.Horizontal)
        self.btns.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.btns.setObjectName("btns")
        self.verticalLayout.addWidget(self.btns)

        self.retranslateUi(GithubAuthenticationDialog)
        self.btns.rejected.connect(GithubAuthenticationDialog.reject)
        self.btns.accepted.connect(GithubAuthenticationDialog.accept)
        QtCore.QMetaObject.connectSlotsByName(GithubAuthenticationDialog)

    def retranslateUi(self, GithubAuthenticationDialog):
        _translate = QtCore.QCoreApplication.translate
        GithubAuthenticationDialog.setWindowTitle(_translate("GithubAuthenticationDialog", "Github Authentication"))
        self.lbl.setText(_translate("GithubAuthenticationDialog", "Instructions to authenticate: ....."))





# -*- coding: utf-8 -*-

################################################################################
#                                                                              #
#   This file is part of the Bibolamazi Project.                               #
#   Copyright (C) 2013 by Philippe Faist                                       #
#   philippe.faist@bluewin.ch                                                  #
#                                                                              #
#   Bibolamazi is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU General Public License as published by       #
#   the Free Software Foundation, either version 3 of the License, or          #
#   (at your option) any later version.                                        #
#                                                                              #
#   Bibolamazi is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#   GNU General Public License for more details.                               #
#                                                                              #
#   You should have received a copy of the GNU General Public License          #
#   along with Bibolamazi.  If not, see <http://www.gnu.org/licenses/>.        #
#                                                                              #
################################################################################

# Py2/Py3 support
from __future__ import unicode_literals, print_function
from past.builtins import basestring
from future.utils import python_2_unicode_compatible, iteritems
from builtins import range
from builtins import str as unicodestr

import sys
import os.path
import logging
logger = logging.getLogger(__name__)

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

import bibolamazi.init
from bibolamazi.core.bibfilter import factory as filters_factory
from bibolamazi.core import main
from bibolamazi.core import butils

from . import githubauthenticationdialog
from . import uiutils
from . import buttontabsmanager
from .uiutils import BlockedSignals
from .qtauto.ui_settingswidget import Ui_SettingsWidget


def get_typewriter_font(widget):
    """
    Return a `QFont` with a typewriter font such as Courier. Use some educated guesses for
    choosing a nice font.
    """
    font = widget.font() # default widget font
    font.setStyleHint(QFont.TypeWriter)
    if sys.platform.startswith("darwin"):
        font.setFamily("Menlo")
        font.setPointSize(12)
    else:
        font.setFamily("Monospace")

    return font


# Model to interface a python dictionary with a `set_at()' method, like our
# `filters.PrependOrderedDict`.
#
class MyOrderedDictModel(QAbstractTableModel):
    def __init__(self, dic, parent=None):
        super(MyOrderedDictModel, self).__init__(parent)
        self._dic = dic

    dicChanged = pyqtSignal()

    def rowCount(self, parent):
        if parent.isValid():
            return 0
        return len(self._dic)

    def columnCount(self, parent):
        if parent.isValid():
            return 0
        return 2

    def data(self, index, role=Qt.DisplayRole):

        col = index.column()
        row = index.row()

        if (row < 0 or row >= len(self._dic)):
            return None

        if role == Qt.BackgroundRole:
            valid = filters_factory.validate_filter_package(str(self._dic.item_at(row)[0]),
                                                            str(self._dic.item_at(row)[1]),
                                                            raise_exception=False)
            if not valid:
                return QBrush(QColor(255,200,200))
            return None

        if (col == 0):
            # argument name
            if (role == Qt.DisplayRole):
                return self._dic.item_at(row)[0]

            # tool-tip documentation
            if (role == Qt.ToolTipRole):
                return self._dic.item_at(row)[0]

            return None
        
        if (col == 1):
            # argument value

            if (role == Qt.DisplayRole or role == Qt.EditRole):
                val = self._dic.item_at(row)[1]
                if (val is None):
                    return None
                return str(val)

            return None

        return None


    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if (orientation == Qt.Vertical):
            return None

        if (section == 0):
            if (role == Qt.DisplayRole):
                return u"Filter Package"
            return None

        if (section == 1):
            if (role == Qt.DisplayRole):
                return u"Path"
            return None

        return None


    def flags(self, index):
        col = index.column()
        row = index.row()

        if col in (0,1,):
            return Qt.ItemIsSelectable | Qt.ItemIsEnabled | Qt.ItemIsEditable

        logger.warning("MyOrderedDictModel.flags(): MyOrderedDictModel.flags(): BAD COLUMN: %d", col)
        return 0
    

    def setData(self, index, value, role=Qt.EditRole):
        
        col = index.column()
        row = index.row()

        if col not in (0,1):
            return False

        if (role != Qt.EditRole):
            return False

        if (row < 0 or row >= len(filters_factory.filterpath)):
            return False

        logger.debug("Got value: %r", value)

        if col == 0:
            self._dic.set_at(row, value, self._dic.item_at(row)[1])
        if col == 1:
            self._dic.set_at(row, self._dic.item_at(row)[0], value)

        self.dataChanged.emit(index, index)

        self.dicChanged.emit()

        return True





def setup_filterpackages_from_settings(s):

    s.beginGroup("BibolamaziCore")

    sval = s.value("filterpath")
    if sval is None:
        fpstr = ''
    else:
        fpstr = str(sval)

    for fp in reversed(fpstr.split(os.pathsep)):
        # if we had 'filters=' from some old version, then replace that by 'bibolamazi.filters='
        if fp == 'filters=':
            fp = 'bibolamazi.filters='

        if not fp: # ignore truly empty entries
            continue

        logger.debug("Adding filter package from settings: fp=\"%s\"", fp)
        main.setup_filterpackage_from_argstr(fp)

    s.endGroup()




class SettingsWidget(QDialog):
    def __init__(self, bibapp=None):
        super(SettingsWidget, self).__init__()

        self.bibapp = bibapp

        self.ui = Ui_SettingsWidget()
        self.ui.setupUi(self)

        if sys.platform.startswith("darwin"):
            self.setWindowFlags(Qt.Sheet)

        # connect buttons as "tabs"
        self.btabs = buttontabsmanager.ButtonTabsManager(self.ui.tabs, self)
        self.btabs.registerButton(self.ui.btnTabGeneral, self.ui.pageGeneral)
        self.btabs.registerButton(self.ui.btnTabLocalRepos, self.ui.pageLocalRepos)
        self.btabs.registerButton(self.ui.btnTabRemoteRepos, self.ui.pageRemoteRepos)

        self.ui.tabs.setCurrentWidget(self.ui.pageGeneral)
        self.ui.btns.setFocus()

        # general

        self.ui.spnMaxRecentFiles.setValue(bibapp.recentFilesList.max_recent_files)

        self.ui.chkHideStartupWindow.setChecked(bibapp.hide_startup_window_on_open_doc)

        # filter packages

        self.fpmodel = MyOrderedDictModel(filters_factory.filterpath)
        self.ui.lstFilterPackages.setModel(self.fpmodel)

        self.ui.lstFilterPackages.selectionModel().selectionChanged.connect(
            self.slot_lstFilterPackages_selectionChanged
            )
        self.slot_lstFilterPackages_selectionChanged()
        self.fpmodel.dicChanged.connect(self.save_settings)

        # update auth status
        self.update_allow_remote_filterpackages()
        self._update_githubauth_guistate()


    def filterpackages_selected_rows(self):
        idxlist = self.ui.lstFilterPackages.selectedIndexes()
        rows = set([i.row() for i in idxlist])
        return list(rows)

    @pyqtSlot()
    def slot_lstFilterPackages_selectionChanged(self):
        if not self.ui.lstFilterPackages.selectionModel().hasSelection():
            self.ui.btnFilterPackageRemove.setEnabled(False)
            self.ui.btnFilterPackageMoveUp.setEnabled(False)
            self.ui.btnFilterPackageMoveDown.setEnabled(False)
            return

        self.ui.btnFilterPackageRemove.setEnabled(True)
        idx = self.filterpackages_selected_rows()
        if len(idx) == 1:
            self.ui.btnFilterPackageMoveUp.setEnabled(idx[0] > 0)
            self.ui.btnFilterPackageMoveDown.setEnabled(idx[0] < len(filters_factory.filterpath)-1)
        else:
            self.ui.btnFilterPackageMoveUp.setEnabled(False)
            self.ui.btnFilterPackageMoveDown.setEnabled(False)
        

    @pyqtSlot()
    def on_btnFilterPackageAdd_clicked(self):

        thedir = str(QFileDialog.getExistingDirectory(self, "Locate Filter Package", str()))

        thekey = os.path.basename(thedir)
        thedir = os.path.dirname(thedir)

        if thekey in filters_factory.filterpath:
            QMessageBox.critical(self, "There is already an existing filter package `%s'!", thekey)

        if thekey and thedir:
            filters_factory.filterpath[thekey] = str(thedir)
            self.ui.lstFilterPackages.reset()
            self.save_settings()

    @pyqtSlot()
    def on_btnFilterPackageRemove_clicked(self):
        idxlist = self.filterpackages_selected_rows()
        if not len(idxlist):
            return

        for idx in idxlist:
            yn = QMessageBox.question(self, "Remove filter package?",
                                      ("Unset filter package %s? The files will not be removed, "
                                      "they will just be ignored by bibolamazi.") % (
                                          filters_factory.filterpath.item_at(idx)[0]
                                          ),
                                      QMessageBox.Yes|QMessageBox.Cancel, QMessageBox.Cancel)
            if yn != QMessageBox.Yes:
                continue
            
            del filters_factory.filterpath[filters_factory.filterpath.item_at(idx)[0]]
            self.ui.lstFilterPackages.reset()
            self.save_settings()

    @pyqtSlot()
    def on_btnFilterPackageMoveUp_clicked(self):
        idxlist = self.filterpackages_selected_rows()
        if len(idxlist) != 1:
            return
        row = idxlist[0]

        if row == 0:
            return

        fpitems = list(filters_factory.filterpath.items())
        filters_factory.filterpath.clear()
        filters_factory.filterpath.update(reversed( fpitems[:row-1] + [fpitems[row], fpitems[row-1]]
                                                    + fpitems[row+1:] ))
        self.ui.lstFilterPackages.reset()
        self.save_settings()
        
    @pyqtSlot()
    def on_btnFilterPackageMoveDown_clicked(self):
        idxlist = self.filterpackages_selected_rows()
        if len(idxlist) != 1:
            return
        row = idxlist[0]

        if row >= len(filters_factory.filterpath)-1:
            return

        fpitems = list(filters_factory.filterpath.items())
        filters_factory.filterpath.clear()
        filters_factory.filterpath.update(reversed( fpitems[:row] + [fpitems[row+1], fpitems[row]]
                                                    + fpitems[row+2:] ))
        self.ui.lstFilterPackages.reset()
        self.save_settings()


    @pyqtSlot(int)
    def on_spnMaxRecentFiles_valueChanged(self, num):
        self.bibapp.recentFilesList.max_recent_files = int(num)


    @pyqtSlot(bool)
    def on_chkHideStartupWindow_toggled(self, on):
        logger.debug("set HideStartupWindowOnOpenDoc = %r", on)
        self.bibapp.setHideStartupWindowOnOpenDoc(on)
        self.save_settings()


    @pyqtSlot(bool)
    def on_chkRemoteAllow_toggled(self, on):
        logger.debug("set remote allowed = %r", on)
        filters_factory.package_provider_manager.saveRemoteAllowedPreference(on)

    def update_allow_remote_filterpackages(self):
        settings = QSettings()
        settings.beginGroup('RemoteFilterPackages')
        # on Windows, Python 3.4 and Qt 5.7, this returns str... so ensure it is a bool
        allow_remote = butils.getbool(settings.value('AllowRemote', False))
        settings.endGroup()
        
        with BlockedSignals(self.ui.chkRemoteAllow):
            self.ui.chkRemoteAllow.setChecked(allow_remote)

        self.ui.grpGithubAuth.setEnabled(allow_remote)

    def _update_githubauth_guistate(self):
        from . import bibolamaziapp
        is_authenticated = False
        if bibolamaziapp.app_filterpackage_providers['github'].getAuthToken():
            is_authenticated = True
            
        if self.ui.chkGithubAuth.isChecked() != is_authenticated:
            with BlockedSignals(self.ui.chkGithubAuth):
                self.ui.chkGithubAuth.setChecked(is_authenticated)

        if is_authenticated:
            self.ui.lblGithubAuthStatus.setText("""\
<span style="color: %(green)s">\N{CHECK MARK} Access token configured</span>
"""%{'green': '#50ff50' if uiutils.is_dark_mode(self) else '#006000'})
        else:
            self.ui.lblGithubAuthStatus.setText("""\
<span style="font-style: italic;"> Currently not authenticated</span>
""") # don't use \N{CROSS MARK}, it's a little too prominent

    @pyqtSlot(bool)
    def on_chkGithubAuth_toggled(self, on):
        logger.debug("Clicked github auth checkbox = %r", on)
        from . import bibolamaziapp
        if not on:
            bibolamaziapp.app_filterpackage_providers['github'].setAuthToken(None)
        else:
            try:
                dlg = githubauthenticationdialog.GithubAuthenticationDialog(self)
                r = dlg.exec_()
                if r == QDialog.Accepted:
                    token = dlg.getAuthToken()
                    logger.debug("Got auth token = [...]%s", token[-4:])
                    bibolamaziapp.set_github_auth_token(token)
            except Exception as e:
                logger.warning("Can't generate access token: %s: %s", e.__class__.__name__, e)
                logger.exception("Exception")
                pass

        self._update_githubauth_guistate()


    @pyqtSlot()
    def save_settings(self):

        logger.debug("settingswidget: Saving settings")

        s = QSettings()

        # max_recent_files is saved along with the recent files themselves

        ### Group "Ui"
        s.beginGroup("Ui")
        s.setValue("HideStartupWindowOnOpenDoc", int(self.ui.chkHideStartupWindow.isChecked()))
        s.endGroup()


        ### Group "BibolamaziCore"
        s.beginGroup("BibolamaziCore")
        s.setValue("filterpath",
                   os.pathsep.join(( "%s=%s"%(k, v if v else "")
                                     for k,v in filters_factory.filterpath.items() ))
                   )
        s.endGroup()

        # reset the filter cache.
        filters_factory.reset_filters_cache()

        s.sync()




# -*- coding: utf-8 -*-

################################################################################
#                                                                              #
#   This file is part of the Bibolamazi Project.                               #
#   Copyright (C) 2013 by Philippe Faist                                       #
#   philippe.faist@bluewin.ch                                                  #
#                                                                              #
#   Bibolamazi is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU General Public License as published by       #
#   the Free Software Foundation, either version 3 of the License, or          #
#   (at your option) any later version.                                        #
#                                                                              #
#   Bibolamazi is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#   GNU General Public License for more details.                               #
#                                                                              #
#   You should have received a copy of the GNU General Public License          #
#   along with Bibolamazi.  If not, see <http://www.gnu.org/licenses/>.        #
#                                                                              #
################################################################################

# Py2/Py3 support
from __future__ import unicode_literals, print_function
from past.builtins import basestring
from future.utils import python_2_unicode_compatible, iteritems
from builtins import range
from builtins import str as unicodestr

import re
import os
import os.path
import logging
logger = logging.getLogger(__name__)


from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

import bibolamazi.init
from bibolamazi.core.helppages import htmlescape, forcewrap_long_lines

from .qtauto.ui_sourcelisteditor import Ui_SourceListEditor

from . import uiutils
from . import helpbrowser



def sanitize_bib_rel_path(fname, ref_dir=None):
    # decide on whether to refer to file in absolute or relative fashion
    if (ref_dir):
        relpath = os.path.relpath(os.path.realpath(fname), os.path.realpath(ref_dir))
        if '..' in relpath.split(os.sep) or '..' in relpath.split(os.altsep):
            # out of scope, so use absolute.
            pass
        else:
            # can easily be referred to by relative path (in same or sub directory), use relative.
            fname = relpath

    homeprefix = os.path.expanduser("~") + os.sep
    if fname.startswith(homeprefix):
        fname = "~/"+fname[len(homeprefix):]

    return fname



class SourceListEditor(QWidget):
    def __init__(self, parent):
        super(SourceListEditor, self).__init__(parent)

        self.ui = Ui_SourceListEditor()
        self.ui.setupUi(self)
        
        self.ui.lblLinkInfo.setVisible(False)

        self.ui.btnAddFavorite.clicked.connect(self.requestAddToFavorites)

        self.sourcelist = []

        #self.ui.lstSources.model().layoutChanged.connect(self.update_stuff_moved)
        #self._is_updating = False

        self._ref_dir = None


    sourceListChanged = pyqtSignal('QStringList')

    requestAddToFavorites = pyqtSignal()
    requestAddSourceList = pyqtSignal()
    

    def sourceList(self):
        if self._erorr_state:
            logger.warning("SourceListEditor's error state is on! don't call sourceList()!")
        return self.sourcelist
        #return [str(self.ui.lstSources.item(i).text())  for i in range(self.ui.lstSources.count())]


    def setRefDir(self, refdir):
        """Sets the \"reference\" directory, which is the directory in which the bibolamazi
        file being edited resides. This is used to decide on whether to refer to a file with
        an absolute or a relative path.
        """
        self._ref_dir = refdir
        

    @pyqtSlot('QStringList')
    def dispSourceListError(self, errormsg):
        self._error_state = True

        self.ui.lbl.setText("<p style=\"color: #800000;\">" +
                            "<b>Parse error:</b> " + htmlescape(errormsg) + "</p>")

    @pyqtSlot('QStringList', bool)
    @pyqtSlot('QStringList')
    def setSourceList(self, newsourcelist, noemit=False):
        self._error_state = False

        newsourcelist = [str(x) for x in list(newsourcelist)]
        
        # don't reset source list if it's the same. In particular, don't emit the changed signal.
        if (newsourcelist == self.sourcelist):
            return

        self.sourcelist = newsourcelist
        self._updated_source_list(noemit=noemit)

    def _updated_source_list(self, noemit=False):
        if self._error_state:
            logger.warning("SourceListEditor's error state is on, don't call _updated_source_list()!")

        def htmlforsrc(i, x):
            code = htmlescape(str(x))
            code += " <a href=\"srcaction:/change/"+str(i)+"\">[\N{HORIZONTAL ELLIPSIS}]</a>"
            if len(self.sourcelist) > 1:
                code += " <a href=\"srcaction:/remove/"+str(i)+"\">[\N{MINUS SIGN}]</a>"
            if i > 0:
                code += " <a href=\"srcaction:/up/"+str(i)+"\">[\N{BLACK UP-POINTING TRIANGLE}]</a>"
            if i < len(self.sourcelist)-1:
                code += " <a href=\"srcaction:/down/"+str(i)+"\">[\N{BLACK DOWN-POINTING TRIANGLE}]</a>"
            return "<p>\N{BLACK RIGHT-POINTING SMALL TRIANGLE} " + code + "</p>"

        if not self.sourcelist:
            self.sourcelist = ['<not set>']

        fileitems = [ htmlforsrc(i,x) for i,x in enumerate(self.sourcelist) ]
        items = fileitems[0]
        if len(fileitems) > 1:
            items += "<p class=\"heading\">Alternatives:</p>" + "".join(fileitems[1:])
        items += ("<p><a href=\"srcaction:/add_alternative/-1\">add alternative source "
                  "location \N{HORIZONTAL ELLIPSIS}</a></p>")
        items += "<p>&nbsp;</p>"

        lbltext = """\
<!DOCTYPE HTML>
<html><head><style type="text/css">
%(basecss)s
.heading { font-style: italic; }
</style></head>
<body>
<p class=\"heading\">Source:</p>
%(sourceitems)s</body></html>
""" % {
    'basecss': helpbrowser.getCssHelpStyle(dark_mode=uiutils.is_dark_mode(self)),
    'sourceitems': items
    }
        #lbltext = ("<ul>" + items + "</ul>")

        self.ui.lbl.setText(lbltext)

        if not noemit:
            self.emitSourceListChanged()

    @pyqtSlot(str)
    def on_lbl_linkHovered(self, link):
        if self._error_state:
            logger.debug("SourceListEditor: Ignoring link interaction while error state is on")
            return

        if not link:
            # link un-hovered
            self.ui.lblLinkInfo.setText("")
            self.ui.lblLinkInfo.setVisible(False)
            return
            
        self.ui.lblLinkInfo.setVisible(True)

        m = re.match(r'^srcaction:/(?P<action>[a-zA-Z0-9_-]+)/(?P<num>[-0-9]+)$', link)
        if m is None:
            logger.warning("Invalid action link: %r", link)
            return
        action = m.group('action')
        i = int(m.group('num'))
        if action == 'change':
            self.ui.lblLinkInfo.setText("Select a new file to replace the current one.")
        elif action == 'remove':
            self.ui.lblLinkInfo.setText("Remove this source location")
        elif action == 'up':
            self.ui.lblLinkInfo.setText("Promote this alternative source location to higher priority")
        elif action == 'down':
            self.ui.lblLinkInfo.setText("Demote this alternative source location to lower priority")
        elif action == 'add_alternative':
            self.ui.lblLinkInfo.setText(
                "Specify an additional bibtex file location where this source can be found.  "
                "This might be useful, for instance, if between different computers or "
                "different collaborators the source bibtex files are stored at different locations."
            )
        else:
            logger.warning("Invalid action link: %r (bad action %r)", link, action)


    @pyqtSlot(str)
    def on_lbl_linkActivated(self, link):

        if self._error_state:
            logger.debug("SourceListEditor: Ignoring link interaction while error state is on")
            return

        m = re.match(r'^srcaction:/(?P<action>[a-zA-Z0-9_-]+)/(?P<num>[-0-9]+)$', link)
        if m is None:
            logger.warning("Invalid action link: %r", link)
            return

        action = m.group('action')
        i = int(m.group('num'))
        if action == 'change':
            self.changeSource(i)
        elif action == 'remove':
            if len(self.sourcelist) > 1:
                del self.sourcelist[i]
                self._updated_source_list()
        elif action == 'up':
            if i > 0:
                self.sourcelist[i], self.sourcelist[i-1] = self.sourcelist[i-1], self.sourcelist[i]
                self._updated_source_list()
        elif action == 'down':
            if i < len(self.sourcelist)-1:
                self.sourcelist[i], self.sourcelist[i+1] = self.sourcelist[i+1], self.sourcelist[i]
                self._updated_source_list()
        elif action == 'add_alternative':
            self.addSourceAlt()
        else:
            logger.warning("Invalid action link: %r (bad action %r)", link, action)

    @pyqtSlot(int, bool)
    @pyqtSlot(int)
    @pyqtSlot()
    def changeSource(self, i=0, noemit=False):
        """
        Prompt user with a file dialog and change the source at index `i`
        """

        if self._error_state:
            logger.warning("SourceListEditor's error state is on, don't call changeSource()!")

        fname = self._get_source_fname()
        if not fname:
            return
        
        self.sourcelist[i] = fname
        self._updated_source_list(noemit=noemit)
        
    @pyqtSlot(bool)
    @pyqtSlot()
    def addSourceAlt(self, noemit=False):

        if self._error_state:
            logger.warning("SourceListEditor's error state is on, don't call addSourceAlt()!")

        fname = self._get_source_fname()
        if not fname:
            return
        
        self.sourcelist.append(fname)
        self._updated_source_list(noemit=noemit)

    def _get_source_fname(self):
        fname, _filter = QFileDialog.getOpenFileName(self, 'Select BibTeX File', str(),
                                                     'BibTeX Files (*.bib);;All Files (*)')
        logger.debug("fname=%r.", fname)
        if (not fname):
            return None

        fname = sanitize_bib_rel_path(fname, ref_dir=self._ref_dir)

        return fname

    
    @pyqtSlot()
    def emitSourceListChanged(self):
        logger.debug("emitting sourceListChanged()!")
        self.sourceListChanged.emit(self.sourcelist)


    @pyqtSlot()
    def on_btnAddSource_clicked(self):
        self.requestAddSourceList.emit()




# -*- coding: utf-8 -*-

################################################################################
#                                                                              #
#   This file is part of the Bibolamazi Project.                               #
#   Copyright (C) 2013 by Philippe Faist                                       #
#   philippe.faist@bluewin.ch                                                  #
#                                                                              #
#   Bibolamazi is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU General Public License as published by       #
#   the Free Software Foundation, either version 3 of the License, or          #
#   (at your option) any later version.                                        #
#                                                                              #
#   Bibolamazi is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#   GNU General Public License for more details.                               #
#                                                                              #
#   You should have received a copy of the GNU General Public License          #
#   along with Bibolamazi.  If not, see <http://www.gnu.org/licenses/>.        #
#                                                                              #
################################################################################

# Py2/Py3 support
from __future__ import unicode_literals, print_function
from past.builtins import basestring
from future.utils import python_2_unicode_compatible, iteritems
from builtins import range
from builtins import str as unicodestr

import logging


from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

#import bibolamazi.init

logger = logging.getLogger(__name__)




class MultiTypesEditorWidget(QWidget):
    def __init__(self, multitypesargtype, parent):
        super(MultiTypesEditorWidget, self).__init__(parent)

        from .filterinstanceeditor import RegisteredArgInputType

        self.setFocusPolicy(Qt.NoFocus)

        self.multitypesargtype = multitypesargtype
        self.typelist = multitypesargtype.typelist

        self.stk = QStackedWidget(self)

        self.typbtn = QToolButton(self)
        self.typbtn.setText("more")
        self.typbtn.setFocusPolicy(Qt.NoFocus)

        self.typmenu = QMenu(self.typbtn)

        factory = QItemEditorFactory()

        self.w = []
        self.wpropnames = []

        self.actiongroup = QActionGroup(self)
        self.actions = []

        for i, T in enumerate(self.typelist):
            t, s = T

            ww = None
            if hasattr(t, 'type_arg_input'):
                rr = RegisteredArgInputType(t, t())
                ww = rr.createWidget(self)
            else:
                qttypeid = QVariant(t()).userType()
                ww = factory.createEditor(qttypeid, self)
            self.stk.addWidget(ww)

            self.w.append(ww)
            self.wpropnames.append(factory.valuePropertyName(qttypeid))

            a = self.typmenu.addAction(s)
            a.setCheckable(True)
            a.setProperty('tindex', i)
            a.triggered.connect(self._set_typ_from_menu)
            self.actiongroup.addAction(a)
            self.actions.append(a)

        self.typbtn.setMenu(self.typmenu)
        self.typbtn.clicked.connect(self.typbtn.showMenu)

        self.tindex = 0

        self.hboxlayout = QHBoxLayout()
        self.hboxlayout.setContentsMargins(1,1,1,1)
        self.hboxlayout.setSpacing(2)
        self.hboxlayout.addWidget(self.stk)
        self.hboxlayout.addWidget(self.typbtn)
        self.hboxlayout.setStretch(0, 99)
        self.hboxlayout.setStretch(1, 0)

        self.setLayout(self.hboxlayout)


    @pyqtSlot()
    def _set_typ_from_menu(self):
        i = self.sender().property('tindex')
        self.tindex = i

        self.stk.setCurrentIndex(i)
        self.w[i].setFocus()


    def setTypeValue(self, typeobj, value):
        from .filterinstanceeditor import RegisteredArgInputType

        for i, T in enumerate(self.typelist):
            t, s = T

            if t is typeobj:
                # found type
                logger.debug("setTypeValue: %r, %r", typeobj, value)

                self.tindex = i
                self.stk.setCurrentIndex(i)
                self.actions[i].setChecked(True)

                if hasattr(t, 'type_arg_input'):
                    rr = RegisteredArgInputType(t, value)
                    rr.setEditorData(self.w[i])
                    self.w[i]
                else:
                    self.w[i].setProperty(self.wpropnames[i], value)

                self.w[i].setFocus()
                return

        raise ValueError("Invalid type: %r"%(typeobj))

    def setValue(self, value):
        t, v = self.multitypesargtype.parse_value_fn(value)
        self.setTypeValue(t, v)
        

    def getTypeValue(self):
        from .filterinstanceeditor import RegisteredArgInputType

        t = self.typelist[self.tindex][0]
        w = self.w[self.tindex]
        if hasattr(t, 'type_arg_input'):
            rr = RegisteredArgInputType(t, t())
            v = rr.valueOf(self.w[self.tindex])
        else:
            v = w.property(self.wpropnames[self.tindex])

        return (t, t(v))





# -*- coding: utf-8 -*-

################################################################################
#                                                                              #
#   This file is part of the Bibolamazi Project.                               #
#   Copyright (C) 2013 by Philippe Faist                                       #
#   philippe.faist@bluewin.ch                                                  #
#                                                                              #
#   Bibolamazi is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU General Public License as published by       #
#   the Free Software Foundation, either version 3 of the License, or          #
#   (at your option) any later version.                                        #
#                                                                              #
#   Bibolamazi is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#   GNU General Public License for more details.                               #
#                                                                              #
#   You should have received a copy of the GNU General Public License          #
#   along with Bibolamazi.  If not, see <http://www.gnu.org/licenses/>.        #
#                                                                              #
################################################################################

# Py2/Py3 support
from __future__ import unicode_literals, print_function
from past.builtins import basestring
from future.utils import python_2_unicode_compatible, iteritems
from builtins import range
from builtins import str as unicodestr
import sys
def to_native_str(x): return x.encode('utf-8') if sys.version_info[0] <= 2 else x
def from_native_str(x): return x.decode('utf-8') if sys.version_info[0] <= 2 else x
from imp import reload

import logging
logger = logging.getLogger(__name__)

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *



class ContextAttributeSetter(object):
    """Give a list of pairs of method and value to set.

    For example:

    >>> with ContextAttributeSetter( (object.isEnabled, object.setEnabled, False), ):
            ...

    will retreive the current state of if the object is enabled with `object.isEnabled()`, then
    will disable the object with `object.setEnabled(False)`. Upon exiting the with block, the
    state is restored to its original state with `object.setEnabled(..)`.

    """

    def __init__(self, *args):
        """Constructor. Does initializations. The \"enter\" statement is done with __enter__().

        Note: the argument are a list of 3-tuples `(get_method, set_method, set_to_value)`.
        """
        super(ContextAttributeSetter, self).__init__()
        self.attribpairs = args
        self.initvals = None

    def __enter__(self):
        self.initvals = []
        for (getm, setm, v) in self.attribpairs:
            self.initvals.append(getm())
            setm(v)
            
        return self

    def __exit__(self, type, value, traceback):
        # clean-up, go in reverse order
        N = len(self.attribpairs)
        for i in range(N):
            (getm, setm, v) = self.attribpairs[N-i-1]
            setm(self.initvals[N-i-1])




class BlockedSignals(ContextAttributeSetter):
    """
    with BlockedSignals(object1, object2, ...):
       # those Qt object's signals are temporarily blocked in this with statement.
    """

    def __init__(self, *args):
        attrlist = [
            (obj.signalsBlocked, obj.blockSignals, True)
            for obj in args
        ]
        super(BlockedSignals, self).__init__( *attrlist )






def is_dark_mode(widget):
    if widget is None:
        return False
    p = widget.palette()
    if p is None:
        return False
    if p.color(QPalette.Active, QPalette.Base).value() < 127:
        # dark Base color -> dark mode
        return True
    return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

################################################################################
#                                                                              #
#   This file is part of the Bibolamazi Project.                               #
#   Copyright (C) 2013 by Philippe Faist                                       #
#   philippe.faist@bluewin.ch                                                  #
#                                                                              #
#   Bibolamazi is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU General Public License as published by       #
#   the Free Software Foundation, either version 3 of the License, or          #
#   (at your option) any later version.                                        #
#                                                                              #
#   Bibolamazi is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#   GNU General Public License for more details.                               #
#                                                                              #
#   You should have received a copy of the GNU General Public License          #
#   along with Bibolamazi.  If not, see <http://www.gnu.org/licenses/>.        #
#                                                                              #
################################################################################


# Py2/Py3 support
from __future__ import unicode_literals, print_function
from past.builtins import basestring
from future.utils import python_2_unicode_compatible, iteritems
from builtins import range
from builtins import str as unicodestr


import sys
import os
import os.path
import re
import logging
import subprocess
import datetime

logger = logging.getLogger(__name__)

import bibolamazi.init

from bibolamazi.core import bibolamazifile
from bibolamazi.core import main
from bibolamazi.core.butils import BibolamaziError
from bibolamazi.core.bibfilter import factory as filters_factory
from bibolamazi.core.bibfilter import argtypes
from bibolamazi.core.bibfilter import pkgprovider, pkgfetcher_github
from bibolamazi.core import version as bibolamaziversion



from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

#print("loaded Qt. executable=", sys.executable, ' meipass=', sys._MEIPASS)

from . import openbibfile
from . import helpbrowser
from . import settingswidget
from . import startupwidget

from .favorites import FavoriteCmdsList
from .newbibolamazifiledialog import NewBibolamazifileDialog



class RecentFile(object):
    def __init__(self, fname, date):
        self.fname = fname
        self.date = date


DEFAULT_MAX_RECENT_FILES = 10


class RecentFilesList(QObject):
    def __init__(self, parent):
        super(RecentFilesList,self).__init__(parent)
        self.max_recent_files = DEFAULT_MAX_RECENT_FILES
        self.files = []

    filesChanged = pyqtSignal()

    def loadFromSettings(self, settings):
        settings.beginGroup("RecentFiles")

        self.files[:] = []

        self.max_recent_files = int(settings.value("max_recent_files", DEFAULT_MAX_RECENT_FILES))
        
        siz = settings.beginReadArray("filelist")
        for i in range(siz):
            settings.setArrayIndex(i)
            fname = settings.value("fname")
            if fname is None: fname = ''
            fname = str(fname)
            date = settings.value("date")
            if date is None: date = QDate.currentDate()
            date = QDate(date).toPyDate()
            self.files.append(RecentFile(fname=fname, date=date))
        settings.endArray()

        settings.endGroup()

        self.filesChanged.emit()

    def saveToSettings(self, settings):
        settings.beginGroup("RecentFiles")

        settings.setValue("max_recent_files", int(self.max_recent_files))

        settings.beginWriteArray("filelist", len(self.files))
        for i,f in enumerate(self.files):
            settings.setArrayIndex(i)
            settings.setValue("fname", str(f.fname))
            settings.setValue("date", QDate(f.date))
        settings.endArray()

        settings.endGroup()

    def addRecentFile(self, fname):
        fp = os.path.realpath(os.path.abspath(fname))
        existing = False
        for i, rf in enumerate(self.files):
            if rf.fname == fp:
                self.files[i].date = datetime.date.today()
                existing = True
        if not existing:
            self.files.insert(0, RecentFile(fp, datetime.date.today()))
        self.files.sort(key=lambda x: x.date, reverse=True) # most recent first
        # drop old files
        if len(self.files) > self.max_recent_files:
            self.files[:] = self.files[:self.max_recent_files]
        self.filesChanged.emit()






class BibolamaziApplication(QApplication):
    def __init__(self, argv):
        super(BibolamaziApplication, self).__init__(argv)
        
        self.setWindowIcon(QIcon(':/pic/bibolamazi_icon.png'))
        self.setApplicationName('Bibolamazi')
        self.setApplicationVersion(bibolamaziversion.version_str)
        self.setOrganizationDomain('org.bibolamazi')
        self.setOrganizationName('Bibolamazi Project')

        if sys.platform.startswith("darwin"):
            self.setQuitOnLastWindowClosed(False)

        # a list of widgets for each open bibolamazi file
        self.openbibfiles = []

        # the help browser
        self.helpbrowser = None

        # the settings widget
        self.settingswidget = None

        # new bibolamazi file dialog
        self.newfiledialog = None

        # favorite commands
        self.favoriteCmdsList = FavoriteCmdsList(parent=self)
        self.favoriteCmdsList.loadFromSettings(QSettings())

        # recent files
        self.recentFilesList = RecentFilesList(parent=self)
        self.recentFilesList.loadFromSettings(QSettings())
        self.recentFilesMenus = []
        self.recentFilesActions = {}
        self.recentFilesList.filesChanged.connect(self.updateRecentFilesMenus)

        # load additional settings
        s = QSettings()
        # -- Ui group
        s.beginGroup("Ui")
        self.hide_startup_window_on_open_doc = \
            s.value("HideStartupWindowOnOpenDoc", sys.platform.startswith('darwin'), type=int)
        s.endGroup()

        # thread in which bibolamazi is run
        self.bibolamazi_thread = openbibfile.global_run_bibolamazi_thread_instance()

        # flag to avoid recursive calls to quit_app()
        self.is_quitting_app = False

        # application actions
        self.myactions = {}
        self.myactions['new'] = QAction("New Bibolamazi File", self)
        self.myactions['new'].triggered.connect(self.createNewFile)
        self.myactions['new'].setShortcut(QKeySequence("Ctrl+N"))
        self.myactions['open'] = QAction("Open Bibolamazi File", self)
        self.myactions['open'].triggered.connect(self.selectOpenFile)
        self.myactions['open'].setShortcut(QKeySequence("Ctrl+O"))
        self.myactions['help'] = QAction("Open Help && Reference Browser", self)
        self.myactions['help'].triggered.connect(self.openHelpBrowser)
        self.myactions['help'].setShortcut(QKeySequence("Ctrl+R"))
        self.myactions['quit'] = QAction("Quit", self)
        self.myactions['quit'].triggered.connect(self.quit_app)
        self.myactions['quit'].setShortcut(QKeySequence("Ctrl+Q"))
        self.myactions['settings'] = QAction("Settings", self)
        self.myactions['settings'].triggered.connect(self.openSettings)
        self.myactions['startupwindow'] = QAction("Show Start-up Window", self)
        
        self.menubar = None
        if sys.platform.startswith('darwin'):
            # Mac OS X
            self.menubar = QMenuBar(None)
            filemenu = self.menubar.addMenu("File")
            filemenu.addAction(self.myactions['new'])
            filemenu.addAction(self.myactions['open'])
            appmenurecentmenu = filemenu.addMenu("Open Recent")
            self.setRecentFilesMenu(appmenurecentmenu)
            filemenu.addAction(self.myactions['settings'])
            filemenu.addSeparator()
            filemenu.addAction(self.myactions['startupwindow'])
            helpmenu = self.menubar.addMenu("Help")
            helpmenu.addAction(self.myactions['help'])

        # create & setup startup widget
        self.startup_widget = startupwidget.StartupWidget(bibapp=self)
        self.startup_widget.showAndRaise()
        self.myactions['startupwindow'].triggered.connect(self.startup_widget.showAndRaise)

        # update recent files menus app-wide
        self.updateRecentFilesMenus()


    recentFilesAvailable = pyqtSignal(bool)


    def openFile(self, fname):
        logger.info("Opening file %r", fname)

        fnamecanon = os.path.realpath(fname)

        logger.debug("canonical file name = %r", fnamecanon)

        for wopen in self.openbibfiles:
            if fnamecanon == wopen.fileName():
                wopen.raise_()
                logger.debug("File %r already open, raising window.", fnamecanon)
                return

        w = openbibfile.OpenBibFile()
        w.setFavoriteCmdsList(self.favoriteCmdsList)
        ok = w.setOpenFile(fname)

        if not ok:
            logger.warning("Failed to open file %s", fname)
            return

        if self.startup_widget.isVisible() and self.hide_startup_window_on_open_doc:
            # Setting designed for Mac OS X.  We show the start-up window, and
            # then we hide it once we open a file or create a new file (to avoid
            # clutter).  You can always show again the start-up window in
            # "File->Show Startup Window".
            self.startup_widget.hide()

        w.show()
        w.raise_()
        w.fileClosed.connect(self.bibFileClosed)

        self.openbibfiles.append(w)
        logger.debug("openbibfile object = %r, self.openbibfiles = %r", w, self.openbibfiles)

        w.requestHelpTopic.connect(self.openHelpTopic)

        if ok:
            self.recentFilesList.addRecentFile(fname)


    def hasOpenFiles(self):
        if self.openbibfiles:
            return True
        return False

        
    @pyqtSlot()
    def selectOpenFile(self):
        openFileDialog = QFileDialog(self.startup_widget, "Open Bibolamazi File", str(),
                                     "Bibolamazi Files (*.bibolamazi.bib);;All Files (*)")

        if sys.platform.startswith('darwin'):
            # NOTE: BUG: OS X' file selection dialog won't understand the
            # .bibolamazi.bib extension (confused by multiple dots). So use Qt's
            # file dialog.
            openFileDialog.setOptions(QFileDialog.DontUseNativeDialog)
        
        openFileDialog.setDefaultSuffix("bibolamazi.bib")
        openFileDialog.setAcceptMode(QFileDialog.AcceptOpen)
        openFileDialog.setFileMode(QFileDialog.AnyFile)
        if not openFileDialog.exec_():
            return
        
        selectedfiles = openFileDialog.selectedFiles()
        if not selectedfiles:
            return

        assert len(selectedfiles) == 1

        fname = selectedfiles[0]
        if fname:
            self.openFile(fname)

        
    def setHideStartupWindowOnOpenDoc(self, on):
        self.hide_startup_window_on_open_doc = bool(on)


    @pyqtSlot()
    def createNewFile(self):
        logger.debug("creating new bibolamazi file")

        # parentwidget = None
        # eventsender = self.sender()
        # if eventsender is not None and isinstance(eventsender, QWidget):
        #     parentwidget = eventsender

        if self.newfiledialog and self.newfiledialog.isVisible():
            self.newfiledialog.raise_()
            return

        self.newfiledialog = NewBibolamazifileDialog(parent=self.startup_widget)
        self.newfiledialog.bibolamaziFileCreated.connect(self.openFile)
        self.newfiledialog.show()
        self.newfiledialog.raise_()




    @pyqtSlot()
    def openHelpBrowser(self):
        if self.helpbrowser is None:
            self.helpbrowser = helpbrowser.HelpBrowser()
        self.helpbrowser.show()
        self.helpbrowser.raise_()

    @pyqtSlot('QString')
    def openHelpTopic(self, path):
        self.openHelpBrowser()
        self.helpbrowser.openHelpTopic(path)

    @pyqtSlot()
    def openSettings(self):
        if self.settingswidget is None:
            self.settingswidget = settingswidget.SettingsWidget(bibapp=self)
            # let the package provider/manager notify the settings widget of changes
            filters_factory.package_provider_manager.settings_widget = self.settingswidget

        self.settingswidget.show()
        self.settingswidget.raise_()



    @pyqtSlot()
    def bibFileClosed(self):
        sender = self.sender()
        if not sender in self.openbibfiles:
            logger.warning("Widget sender %r of fileClosed() not in our openbibfiles list %r!!",
                           sender, self.openbibfiles)
            return
        logger.debug("file is closed (sender=%r).", sender)
        self.openbibfiles.remove(sender)

        # on Mac OS X, if the last window was closed, then show the start-up
        # window:
        if not self.startup_widget.isVisible():
            if not len(self.openbibfiles):
                #    QTimer.singleShot(100, self.startup_widget.showAndRaise)
                self.startup_widget.showAndRaise()


    @pyqtSlot()
    def quit_app(self):
        if self.is_quitting_app:
            return
        self.is_quitting_app = True
        try:

            logger.info("App quit")

            settings = QSettings()
            self.favoriteCmdsList.saveToSettings(settings)
            self.recentFilesList.saveToSettings(settings)

            # close open bib files one by one
            for i in range(len(self.openbibfiles)):
                w = self.openbibfiles[i]
                ans = w.close() # this will call bibFileClosed and remove the window
                                # from the openbibfiles list
                if not ans:
                    # if the widget cancels the close, then abort
                    event.ignore()
                    return False

            # at this point, quit is confirmed and we shut down everything

            while self.bibolamazi_thread.busy:
                QThread.currentThread().usleep(100000) # 0.1 seconds
            self.bibolamazi_thread.doQuit()

            if self.helpbrowser:
                self.helpbrowser.close()

            self.quit()

        finally:
            self.is_quitting_app = False



    def event(self, event):
        if event.type() == QEvent.FileOpen:
            logger.info("Opening file %s", event.file())
            self.openFile(event.file())
            return True

        if event.type() == QEvent.Close:
            # Cmd-Q
            logger.debug("Intercepting Cmd-Q, quitting gracefully...")
            QTimer.singleShot(0, self.quit_app)
            event.ignore()
            return True
            
        return super(BibolamaziApplication, self).event(event)



    def setRecentFilesMenu(self, menu):
        """
        Register `menu` as being a "recent-files" menu.  The menu will be populated
        with recent files and updated whenever the recent files list changes.
        """
        self.recentFilesMenus.append(menu)


    @pyqtSlot()
    def updateRecentFilesMenus(self):

        files = self.recentFilesList.files

        for menu in self.recentFilesMenus:
            menu.clear()

        if not files:
            self.recentFilesAvailable.emit(False)
            return

        self.recentFilesAvailable.emit(True)

        # file name title to show in menu -- add some path information in case
        # file names are the same
        def fpathitems(fpath, num=3):
            p, f = os.path.split(fpath)
            if not f: return []
            if num <= 1:
                return [f]
            return fpathitems(p, num-1) + [f]
        def ftitle(fname):
            fpath, fn = os.path.split(fname)
            return fn + " (" + os.sep.join(fpathitems(fpath)) + ")"

        for rf in files:
            if rf.fname not in self.recentFilesActions:
                a = QAction(ftitle(rf.fname), self)
                a.triggered.connect(self._openRecentFileSlot)
                a.setProperty("recentFileName", rf.fname)
                self.recentFilesActions[rf.fname] = a

            for menu in self.recentFilesMenus:
                menu.addAction(self.recentFilesActions[rf.fname])

    @pyqtSlot()
    def _openRecentFileSlot(self):
        action = self.sender()
        fname = action.property("recentFileName")
        self.openFile(fname)


#
class AppPackageProviderManager(pkgprovider.PackageProviderManager):
    def __init__(self):
        super(AppPackageProviderManager, self).__init__()
        settings = QSettings()
        settings.beginGroup('RemoteFilterPackages')
        self.prompted_for_remote = settings.value('PromptedForRemote', False)
        self.allow_remote = settings.value('AllowRemote', False)
        settings.endGroup()

        self.settings_widget = None

    def saveRemoteAllowedPreference(self, allow_remote):

        self.allow_remote = allow_remote

        self.prompted_for_remote = True
        settings = QSettings()
        settings.beginGroup('RemoteFilterPackages')
        settings.setValue('PromptedForRemote', True)
        settings.setValue('AllowRemote', self.allow_remote)
        settings.endGroup()
        settings.sync()

        if self.settings_widget:
            self.settings_widget.update_allow_remote_filterpackages()

        if self.allow_remote:
            logger.warning("Allowing remote filter packages for future sessions. "
                           "You can change this in the settings.")
        

    def remoteAllowed(self):
        if not self.prompted_for_remote:
            # ask for remote
            r = QMessageBox.warning(
                None,
                "Allow remote filter packages",
                "Filter packages are python scripts that can execute arbitrary code. "
                "You should only run filters from sources you trust.  Do you want to "
                "enable automatically downloading remote packages when a remote package "
                "is specified?  (You can change this later in the settings window.)",
                QMessageBox.Yes | QMessageBox.No
            )

            allow_remote = ( r == QMessageBox.Yes )

            self.saveRemoteAllowedPreference(allow_remote)

        elif not self.allow_remote:
            logger.warning("Remote filter packages have been disabled. To enable remote "
                           "filter packages, you need to enable the corresponding option "
                           "in the settings.")

        return self.allow_remote



app_filterpackage_providers = {}

def load_filterpackage_providers(app):
    settings = QSettings()

    # first, create our package provider manager.
    filters_factory.package_provider_manager = AppPackageProviderManager()

    settings.beginGroup('RemoteFilterPackages')
    github_auth_token = settings.value('GithubAuthToken', '').strip()
    settings.endGroup()

    app_filterpackage_providers['github'] = \
        pkgfetcher_github.GithubPackageProvider(github_auth_token if github_auth_token else None)

    filters_factory.package_provider_manager.registerPackageProvider(
        app_filterpackage_providers['github']
    )

def set_github_auth_token(github_auth_token):
    app_filterpackage_providers['github'].setAuthToken(github_auth_token)

    settings = QSettings()

    settings.beginGroup('RemoteFilterPackages')
    settings.setValue('GithubAuthToken', github_auth_token)
    settings.endGroup()

    settings.sync()


def run_app(argv):
    
    logger.debug("starting application")

    app = BibolamaziApplication(argv)


    # set up the filter package providers
    #
    load_filterpackage_providers(app)


    try:
        # load filter packages from environment ...
        main.setup_filterpackages_from_env()
        # ... and from settings.
        settingswidget.setup_filterpackages_from_settings(QSettings())
    except (filters_factory.NoSuchFilter, filters_factory.NoSuchFilterPackage, BibolamaziError):
        QMessageBox.warning(None, "Filter packages error",
                            "An error was detected in the filter packages configuration. "
                            "Please edit your settings.")
        pass


    args = app.arguments()
    _rxscript = re.compile('\.(py[co]?|exe)$', flags=re.IGNORECASE)
    for k in range(1,len(args)): # skip program name == argv[0]
        fn = str(args[k])
        if _rxscript.search(fn):
            # our own script, bug on windows?
            logger.debug("skipping own arg: %s", fn)
            continue
        
        logger.debug("opening arg: %s", fn)
        app.openFile(fn)


    sys.exit(app.exec_())
    



#!/usr/bin/env python
# -*- coding: utf-8 -*-

################################################################################
#                                                                              #
#   This file is part of the Bibolamazi Project.                               #
#   Copyright (C) 2013 by Philippe Faist                                       #
#   philippe.faist@bluewin.ch                                                  #
#                                                                              #
#   Bibolamazi is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU General Public License as published by       #
#   the Free Software Foundation, either version 3 of the License, or          #
#   (at your option) any later version.                                        #
#                                                                              #
#   Bibolamazi is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#   GNU General Public License for more details.                               #
#                                                                              #
#   You should have received a copy of the GNU General Public License          #
#   along with Bibolamazi.  If not, see <http://www.gnu.org/licenses/>.        #
#                                                                              #
################################################################################


# Py2/Py3 support
from __future__ import unicode_literals, print_function
from past.builtins import basestring
from future.utils import python_2_unicode_compatible, iteritems
from builtins import range
from builtins import str as unicodestr

import sys
import os
import os.path
import re
import logging
import subprocess
import datetime

#if sys.platform.startswith("win"):
#    # why do we need this??!??!?!
#    logging.basicConfig(level=logging.DEBUG)

import bibolamazi.init

# set up basic logging
from bibolamazi.core import blogger

from bibolamazi.core import bibolamazifile
#from bibolamazi.core import main as bibolamazi_main
from bibolamazi.core.butils import BibolamaziError
from bibolamazi.core.bibfilter import factory as filters_factory
from bibolamazi.core.bibfilter import argtypes
from bibolamazi.core import version as bibolamaziversion

# Important: Do not import Qt5 yet
#
# We need to set some environment variables, etc., to make sure Qt can find all
# its libraries and plugins well when using PyInstaller.



#print("Bibolamazi: loading main script")

logger = logging.getLogger(__name__)


def setup_qt5_plugins_path():
    try:
        basePath = sys._MEIPASS
    except Exception:
        return

    qt_platform_plugins_path = os.path.join(basePath, "PyQt5", "Qt", "plugins")

    logger.debug("Setting QT_QPA_PLATFORM_PLUGIN_PATH to %r", qt_platform_plugins_path)

    os.environ['QT_QPA_PLATFORM_PLUGIN_PATH'] = qt_platform_plugins_path



def exception_handler(etype, evalue, tb):
    if etype is KeyboardInterrupt:
        print("*** interrupt (ignoring)")
        return

    import traceback
    logger.warning("Internal Error: Uncaught Python Exception. Ignoring and hoping for the best.\n%s",
                   "".join(traceback.format_exception(etype, evalue, tb)))

# override Python's exception handler, 
sys.excepthook = exception_handler




def main():


    blogger.setup_simple_console_logging(level=logging.DEBUG)

    argv = list(sys.argv)

    # default level: set to root logger.  May be set externally via environment variable
    # (e.g. for debugging) or with --verbose

    log_level = logging.INFO
    if '--verbose' in argv:
        argv.remove('--verbose')
        log_level = logging.DEBUG

    if 'BIBOLAMAZI_LOG_LEVEL' in os.environ and os.environ['BIBOLAMAZI_LOG_LEVEL']:
        logging.getLogger().setLevel(argtypes.LogLevel(os.environ['BIBOLAMAZI_LOG_LEVEL']).levelno)
    else:
        logging.getLogger().setLevel(log_level)

    logger.debug("main: log level set to %r", log_level)


    # setup Qt paths!
    setup_qt5_plugins_path()



    # ## Seems we still need this for pyinstaller, I'm not sure why....
    #
    # load precompiled filters, if we've got any
    try:
        import bibolamazi_compiled_filter_list as pc
        filters_factory.load_precompiled_filters('bibolamazi.filters', dict([
            (fname, pc.__dict__[fname])  for fname in pc.filter_list
            ]))
    except ImportError:
        pass


    # note: package providers will be loaded after loading the Qt application:

    from . import bibolamaziapp

    return bibolamaziapp.run_app(argv)


if __name__ == '__main__':

    main()




# -*- coding: utf-8 -*-

################################################################################
#                                                                              #
#   This file is part of the Bibolamazi Project.                               #
#   Copyright (C) 2013 by Philippe Faist                                       #
#   philippe.faist@bluewin.ch                                                  #
#                                                                              #
#   Bibolamazi is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU General Public License as published by       #
#   the Free Software Foundation, either version 3 of the License, or          #
#   (at your option) any later version.                                        #
#                                                                              #
#   Bibolamazi is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#   GNU General Public License for more details.                               #
#                                                                              #
#   You should have received a copy of the GNU General Public License          #
#   along with Bibolamazi.  If not, see <http://www.gnu.org/licenses/>.        #
#                                                                              #
################################################################################

# Py2/Py3 support
from __future__ import unicode_literals, print_function
from past.builtins import basestring
from future.utils import python_2_unicode_compatible, iteritems
from builtins import range
from builtins import str as unicodestr

import logging

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

import bibolamazi.init
from .qtauto.ui_searchwidget import Ui_SearchWidget

from . import uiutils

logger = logging.getLogger(__name__)


class SearchWidget(QWidget):
    def __init__(self, parent):
        super(SearchWidget, self).__init__(parent)

        self.u = Ui_SearchWidget()
        self.u.setupUi(self)

        self.default_palette = QPalette(self.u.txt.palette())
        self.found_palette = QPalette(self.u.txt.palette())
        self.notfound_palette = QPalette(self.u.txt.palette())

        self.found_palette.setBrush(QPalette.Base, QColor(180,255,180,80))
        self.notfound_palette.setBrush(QPalette.Base, QColor(255,180,180,80))

        self.u.btnDone.clicked.connect(self.hide)
        self.u.btnDone.clicked.connect(self.searchDone)

        self.u.txt.returnPressed.connect(self.findNext)
        self.u.txt.installEventFilter(self)

        self.setVisible(False)


    def eventFilter(self, object, event):
        if object is self.u.txt:
            if event.type() == QEvent.KeyPress and hasattr(event, 'matches'):
                if event.key() == Qt.Key_Escape:
                    self.searchDone.emit()
                    self.setVisible(False)
                    return True
                if event.matches(QKeySequence.FindNext):
                    self.findNext.emit()
                    return True
                if event.matches(QKeySequence.FindPrevious):
                    self.findPrev.emit()
                    return True

        return super(SearchWidget, self).eventFilter(object, event)
        


    def searchString(self):
        return self.u.txt.text()

    doSearch = pyqtSignal(str)
    searchDone = pyqtSignal()

    findNext = pyqtSignal()#str)
    findPrev = pyqtSignal()#str)

    @pyqtSlot()
    @pyqtSlot(str)
    def showSearch(self, searchstr=None):
        self.u.txt.setText(searchstr)
        self.u.txt.setPalette(self.default_palette)
        self.setVisible(True)
        self.u.txt.setFocus()

    @pyqtSlot()
    def showFound(self):
        self.u.txt.setPalette(self.found_palette)

    @pyqtSlot()
    def showNotFound(self):
        self.u.txt.setPalette(self.notfound_palette)

    @pyqtSlot(str)
    def on_txt_textChanged(self, s):
        self.doSearch.emit(s)

    @pyqtSlot()
    def on_btnNext_clicked(self):
        self.findNext.emit()#self.searchString())

    @pyqtSlot()
    def on_btnPrev_clicked(self):
        self.findPrev.emit()#self.searchString())

#


class SearchTextEditManager(QObject):
    def __init__(self, searchwidget, textedit):
        super(SearchTextEditManager, self).__init__()

        self.textedit = textedit
        self.textedit.installEventFilter(self)

        self.searchwidget = searchwidget

        self.pos = -1
        self.search_matches = None

        highlightcolor = None
        if uiutils.is_dark_mode(self.textedit):
            # dark base color, use corresponding dark hightlight color
            highlightcolor = QColor(130,130,80)
        else:
            # light base color, use light highlight color
            highlightcolor = QColor(255,255,128)
        

        self.highlight_format = QTextCharFormat()
        self.highlight_format.setBackground(highlightcolor)
        self.highlight_format.setFontUnderline(True)
        self.highlight_format.setUnderlineColor(highlightcolor) #QColor(180,180,80))

        self.cur_highlight_format = QTextCharFormat()
        self.cur_highlight_format.setBackground(QColor(220,120,0))
        self.cur_highlight_format.setForeground(QColor(255,255,255))
        self.cur_highlight_format.setFontUnderline(True)
        self.cur_highlight_format.setUnderlineColor(QColor(128,128,128))

        self.searchwidget.doSearch.connect(self.doSearch)
        self.searchwidget.findNext.connect(self.findNext)
        self.searchwidget.findPrev.connect(self.findPrev)
        self.searchwidget.searchDone.connect(self.searchDone)
        

    def eventFilter(self, object, event):
        if object is self.textedit:
            if event.type() == QEvent.KeyPress and hasattr(event, 'matches'):
                if event.matches(QKeySequence.Find):
                    self.showSearch()
                    return True
                if event.matches(QKeySequence.FindNext):
                    self.showSearch(self.searchwidget.searchString())
                    self.findNext()
                    return True
                if event.matches(QKeySequence.FindPrevious):
                    self.showSearch(self.searchwidget.searchString())
                    self.findPrev()
                    return True

        return super(SearchTextEditManager, self).eventFilter(object, event)

    @pyqtSlot()
    @pyqtSlot(str)
    def showSearch(self, s=''):
        self.searchwidget.showSearch(s)
        self.pos = self.textedit.textCursor().position() - 1

    @pyqtSlot(str)
    def doSearch(self, s):
        self.search_matches = []

        doc = self.textedit.document()

        c = QTextCursor()
        while True:
            c = doc.find(s, c)
            if c.isNull():
                break
            self.search_matches.append(c)

        self._set_extra_selections()

        if len(self.search_matches):
            self.searchwidget.showFound()
        else:
            self.searchwidget.showNotFound()

    def _set_extra_selections(self):
        extrasels = []
        for m in self.search_matches:
            sel = QTextEdit.ExtraSelection()
            sel.cursor = m
            sel.format = (self.cur_highlight_format if m.position() == self.pos
                          else self.highlight_format)
            extrasels.append(sel)

        self.textedit.setExtraSelections(extrasels)

    @pyqtSlot()
    def searchDone(self):
        self.textedit.setExtraSelections([])
        self.search_matches = None
        self.pos = -1
        self.textedit.setFocus() # return focus to text edit


    def _start_cursor(self):
        c = QTextCursor(self.textedit.document())
        c.movePosition(QTextCursor.Start)
        return c

    def _nextmatch(self):
        if self.search_matches is None:
            self.doSearch(self.searchwidget.searchString())
        return next( ( cc for cc in self.search_matches if cc.position() > self.pos ),
                     QTextCursor() )

    def _prevmatch(self):
        if self.search_matches is None:
            self.doSearch(self.searchwidget.searchString())
        return next( (cc for cc in reversed(self.search_matches)
                      if self.pos < 0 or cc.position() < self.pos),
                     QTextCursor() )

    def _show_match(self, c):
        self.searchwidget.showFound()
        self.pos = c.position()
        self._set_extra_selections()
        
        c2 = QTextCursor(c)
        c2.clearSelection()
        self.textedit.setTextCursor(c2)
        self.textedit.ensureCursorVisible()

    def _show_notfound(self):
        self.textedit.setTextCursor(self._start_cursor())
        self.searchwidget.showNotFound()
        self.pos = -1
        self._set_extra_selections()

    @pyqtSlot()
    def findNext(self):
        cnext = self._nextmatch()
        if cnext.isNull():
            self._show_notfound()
            return

        self._show_match(cnext)

    @pyqtSlot()
    def findPrev(self):
        cprev = self._prevmatch()
        if cprev.isNull():
            self._show_notfound()
            return

        self._show_match(cprev)



# -*- coding: utf-8 -*-

################################################################################
#                                                                              #
#   This file is part of the Bibolamazi Project.                               #
#   Copyright (C) 2013 by Philippe Faist                                       #
#   philippe.faist@bluewin.ch                                                  #
#                                                                              #
#   Bibolamazi is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU General Public License as published by       #
#   the Free Software Foundation, either version 3 of the License, or          #
#   (at your option) any later version.                                        #
#                                                                              #
#   Bibolamazi is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#   GNU General Public License for more details.                               #
#                                                                              #
#   You should have received a copy of the GNU General Public License          #
#   along with Bibolamazi.  If not, see <http://www.gnu.org/licenses/>.        #
#                                                                              #
################################################################################

# Py2/Py3 support
from __future__ import unicode_literals, print_function
from past.builtins import basestring
from future.utils import python_2_unicode_compatible, iteritems
from builtins import range
from builtins import str as unicodestr

import re
from collections import namedtuple

import bibolamazi.init
# bibolamazi filters
from bibolamazi.core.bibfilter import factory as filters_factory
from bibolamazi.core.bibolamazifile import BIBOLAMAZIFILE_COMMANDS

from PyQt5.QtCore import *
from PyQt5.QtGui import *



rxsrc = re.compile(r'^\s*(?P<src>src:)', re.MULTILINE)
rxfilter = re.compile(r'^\s*(?P<filter>filter:)\s+(?P<filtername>[-:\w]+)', re.MULTILINE)
rxcmdother = re.compile(r'^\s*(?P<cmd>('
                        + '|'.join([x for x in BIBOLAMAZIFILE_COMMANDS if x not in ['src', 'filter']])
                        + r'):)', re.MULTILINE)
rxcomment = re.compile(r'^\s*%%.*$', re.MULTILINE)
_rx_not_odd_num_backslashes = r'(((?<=[^\\])|^)(\\\\)*)';
rxstring1 = re.compile(_rx_not_odd_num_backslashes+r'(?P<str>\"([^"\\]|\\\\|\\\")*\")', re.MULTILINE)
rxstring2 = re.compile(_rx_not_odd_num_backslashes+r"(?P<str>\'[^']*\')", re.MULTILINE)


class BibolamaziConfigSyntaxHighlighter(QSyntaxHighlighter):
    def __init__(self, parent=None):
        super(BibolamaziConfigSyntaxHighlighter, self).__init__(parent)

        self.fmt_src = QTextCharFormat()
        self.fmt_src.setFontWeight(QFont.Bold)
        self.fmt_src.setForeground(QColor(0, 180, 180))

        self.fmt_filter = QTextCharFormat()
        self.fmt_filter.setFontWeight(QFont.Bold)
        self.fmt_filter.setForeground(QColor(200,0,0))

        self.fmt_cmdother = QTextCharFormat()
        self.fmt_cmdother.setFontWeight(QFont.Bold)

        self.fmt_filtername = QTextCharFormat()
        self.fmt_filtername.setForeground(QColor(100, 100, 200))

        self.fmt_filtername_nonex = QTextCharFormat(self.fmt_filtername)
        self.fmt_filtername_nonex.setUnderlineStyle(QTextCharFormat.SpellCheckUnderline)

        self.fmt_comment = QTextCharFormat()
        self.fmt_comment.setForeground(QColor(120,140,120))
        self.fmt_comment.setFontItalic(True)

        self.fmt_string = QTextCharFormat()
        self.fmt_string.setForeground(QColor(0,180,0))


    def highlightBlock(self, text):

        #pcache = BibConfigParsingCache()
        
        blockno = self.currentBlock().blockNumber()

        for m in rxsrc.finditer(text):
            self.setFormat(m.start('src'), len(m.group('src')), self.fmt_src)
            #pcache.add_sourcelist(line=blockno)

        for m in rxfilter.finditer(text):
            self.setFormat(m.start('filter'), len(m.group('filter')), self.fmt_filter)
            fmtname = self.fmt_filtername
            try:
                # try to load the filter module to see if it exists
                filtmodule = filters_factory.FilterInfo(m.group('filtername'))
            except Exception:
                fmtname = self.fmt_filtername_nonex

            #pcache.add_filter(line=blockno, filtername=m.group('filtername'))
                
            self.setFormat(m.start('filtername'), len(m.group('filtername')), fmtname)

        for m in rxcmdother.finditer(text):
            self.setFormat(m.start('cmd'), len(m.group('cmd')), self.fmt_cmdother)

        for m in rxstring1.finditer(text):
            self.setFormat(m.start('str'), len(m.group('str')), self.fmt_string)
        for m in rxstring2.finditer(text):
            self.setFormat(m.start('str'), len(m.group('str')), self.fmt_string)

        for m in rxcomment.finditer(text):
            self.setFormat(m.start(), len(m.group()), self.fmt_comment)

        #self.setCurrentBlockState(0)
        #self.setCurrentBlockUserData(pcache)




# -*- coding: utf-8 -*-

################################################################################
#                                                                              #
#   This file is part of the Bibolamazi Project.                               #
#   Copyright (C) 2018 by Philippe Faist                                       #
#   philippe.faist@bluewin.ch                                                  #
#                                                                              #
#   Bibolamazi is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU General Public License as published by       #
#   the Free Software Foundation, either version 3 of the License, or          #
#   (at your option) any later version.                                        #
#                                                                              #
#   Bibolamazi is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#   GNU General Public License for more details.                               #
#                                                                              #
#   You should have received a copy of the GNU General Public License          #
#   along with Bibolamazi.  If not, see <http://www.gnu.org/licenses/>.        #
#                                                                              #
################################################################################

# Py2/Py3 support
from __future__ import unicode_literals, print_function
from past.builtins import basestring
from future.utils import python_2_unicode_compatible, iteritems
from builtins import range
from builtins import str as unicodestr

import os
import os.path
import re
import logging
import textwrap
from html import escape as htmlescape

import bibolamazi.init

from bibolamazi.core.bibfilter import factory as filters_factory

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

from .qtauto.ui_filterpackagepatheditor import Ui_FilterPackagePathEditor
from .sourcelisteditor import sanitize_bib_rel_path
from . import githubreposelector

logger = logging.getLogger(__name__)



# ------------------------------------------------------------------------------



class FilterPackagePathEditor(QWidget):
    def __init__(self, parent):
        super(FilterPackagePathEditor, self).__init__(parent)

        self.ui = Ui_FilterPackagePathEditor()
        self.ui.setupUi(self)

        self.ref_dir = None

    filterPackagePathChanged = pyqtSignal(str)

    def setRefDir(self, ref_dir):
        self.ref_dir = ref_dir

    @pyqtSlot(str, str)
    def setFilterPackageInfo(self, arg):
        fpspec = filters_factory.FilterPackageSpec(arg)
        if fpspec.is_url:
            self.ui.lblInfo.setText("Location: <b>{}</b>".format(htmlescape(str(fpspec.url))))
        else:
            self.ui.lblInfo.setText("<b>{}:</b> {}".format(htmlescape(str(fpspec.fpname)),
                                                           htmlescape(str(fpspec.fpdir))))

    @pyqtSlot(str)
    def setFilterPackageError(self, errmsg):
        self.ui.lblInfo.setText("<span style=\"color: #800000\">{}</span>".format(htmlescape(errmsg)))


    @pyqtSlot()
    def on_btnSetLocalPackage_clicked(self):

        fpath = QFileDialog.getExistingDirectory(self, "Locate Filter Package", str())

        logger.debug("User selected fpath = %r", fpath)

        if not fpath:
            return

        fpath = sanitize_bib_rel_path(fpath, ref_dir=self.ref_dir)
        self.filterPackagePathChanged.emit(fpath)

    @pyqtSlot()
    def on_btnSetGithubRepo_clicked(self):

        gh = githubreposelector.GithubRepoSelector(self)
        
        r = gh.exec_()
        if r != QDialog.Accepted:
            return

        url = gh.getFilterPackageUrl()

        self.filterPackagePathChanged.emit(url)




# -*- coding: utf-8 -*-

################################################################################
#                                                                              #
#   This file is part of the Bibolamazi Project.                               #
#   Copyright (C) 2013 by Philippe Faist                                       #
#   philippe.faist@bluewin.ch                                                  #
#                                                                              #
#   Bibolamazi is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU General Public License as published by       #
#   the Free Software Foundation, either version 3 of the License, or          #
#   (at your option) any later version.                                        #
#                                                                              #
#   Bibolamazi is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#   GNU General Public License for more details.                               #
#                                                                              #
#   You should have received a copy of the GNU General Public License          #
#   along with Bibolamazi.  If not, see <http://www.gnu.org/licenses/>.        #
#                                                                              #
################################################################################

# Py2/Py3 support
from __future__ import unicode_literals, print_function
from past.builtins import basestring
from future.utils import python_2_unicode_compatible, iteritems
from builtins import range
from builtins import str as unicodestr
from future.standard_library import install_aliases
install_aliases()

import re
import sys
import logging
import os.path
from collections import OrderedDict

from urllib.parse import urlsplit

import markdown2

import bibolamazi.init
from bibolamazi.core import main as bibolamazimain
from bibolamazi.core import blogger
from bibolamazi.core.blogger import logger
from bibolamazi.core import butils
from bibolamazi.core import argparseactions
from bibolamazi.core.bibfilter import factory as filters_factory
from bibolamazi.core import helppages
from bibolamazi.core.helppages import htmlescape, forcewrap_long_lines

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

from . import uiutils
from . import settingswidget
from . import searchwidget

from .qtauto.ui_helpbrowser import Ui_HelpBrowser

logger = logging.getLogger(__name__)





def getCssHelpStyle(fontsize='medium', fontsize_big='large',
                    fontsize_code='medium', fontsize_small='small',
                    dark_mode=False):
    return (
        _HTML_CSS % {'fontsize': fontsize,
                     'fontsize_big': fontsize_big,
                     'fontsize_code': fontsize_code,
                     'fontsize_small': fontsize_small}
        + ( _HTML_CSS_COLORS if not dark_mode else _HTML_CSS_COLORS_DARK )
    )

TABLE_WIDTH = 550 # px

_HTML_CSS = '''
.content {
}
p {
  margin-top: 0pt;
  margin-bottom: 1em;
}
p, a, li, span {
  font-size: %(fontsize)s;
}
em, i {
  font-size: %(fontsize)s;
  font-style: italic;
}
strong, b {
  font-size: %(fontsize)s;
  font-style: bold;
}
code {
  font-size: %(fontsize_code)s;
}
.code-meta {
  font-style: italic;
}
.small {
  font-size: %(fontsize_small)s;
}
pre {
  margin-left: 12px;
  font-size: %(fontsize_code)s;
}
pre.txtcontent { margin-left: 0px; }
a { text-decoration: none }

.biglink{
  text-decoration: underline;
  font-size: %(fontsize_big)s;
}

li {
  margin-bottom: 0.35em;
}
dt {
  font-weight: bold;
  margin-left: 25px;
}
dd {
  margin-left: 50px;
}
table {
  margin-top: 1em;
  margin-bottom: 1em;
}
th {
  font-weight: bold;
  text-align: left;
  padding-right: 5px;
}
td {
  font-size: %(fontsize)s;
  padding-bottom: 0.3em;
}
td.indent {
  padding-left: 2em;
}
td p.inner {
  margin-bottom: 0.1em;
}
'''

_HTML_CSS_COLORS = '''
html, body {
  background-color: #ffffff;
  color: #303030;
}
.code-meta {
  color: #80b0b0;
}
.shadow {
  color: #a0a0a0;
}
a { color: #0000a0; }
th { color: #600030; }
'''

_HTML_CSS_COLORS_DARK = '''
html, body {
  background-color: #202020;
  color: #d0d0d0;
}
.code-meta {
  color: #a0d7d7;
}
.shadow {
  color: #808080;
}
a { color: #80a0ff; }
th { color: #c04080; }
'''


def wrapInHtmlContentContainer(htmlcontent, dark_mode=False, width=None):
    if width is None:
        width = TABLE_WIDTH

    html_top = ("<html><head><style type=\"text/css\">" +
                getCssHelpStyle(dark_mode=dark_mode) +
                "</style></head>" +
                "<body>")
    html_bottom = "</body></html>"

    return (html_top +
            "<table width=\""+str(width)+"\" style=\"margin-left:15px\">" +
            "<tr><td class=\"content\">" +
            htmlcontent +
            "</td></tr></table>" +
            html_bottom)





class HelpTopicPageWidget(QWidget):
    def __init__(self, helptopicpage, helpbrowser, parent):
        super(HelpTopicPageWidget, self).__init__(parent)

        self.lyt = QVBoxLayout(self)
        self.lyt.setContentsMargins(0,0,0,0)
        self.lyt.setSpacing(5)

        self.searchwidget = searchwidget.SearchWidget(self)
        self.lyt.addWidget(self.searchwidget)

        self.tb = QTextBrowser(self)
        self.lyt.addWidget(self.tb)

        self.setLayout(self.lyt)

        self.searchmgr = searchwidget.SearchTextEditManager(self.searchwidget, self.tb)

        font = self.tb.font()
        font.setPointSize(QFontInfo(font).pointSize()+1)
        self.tb.setFont(font)

        self.tb.setOpenLinks(False)
        self.tb.anchorClicked.connect(helpbrowser.openHelpTopicUrl)

        dark_mode = uiutils.is_dark_mode(self)

        htmlfragment = helptopicpage.contentAsHtmlFragment()
        logger.longdebug("Help page html fragment = \n%s", htmlfragment)
        html = wrapInHtmlContentContainer(htmlfragment, dark_mode=dark_mode)
        self.tb.setHtml(html)
        
        self.setProperty("HelpTabTitle", helptopicpage.title())
        if helptopicpage.desc:
            self.setProperty("HelpTabToolTip", helptopicpage.desc())


    def scrollToAnchor(self, aname):
        self.tb.scrollToAnchor(aname)


class TabAlreadyOpen(Exception):
    def __init__(self, widget):
        super(TabAlreadyOpen, self).__init__()
        self.widget = widget



_rx_scheme_frag = re.compile(
    r'^((?P<scheme>[A-Za-z0-9_.-]+):)?(?P<rest>.*)(\#(?P<fragment>[a-zA-Z0-9%_.-]+))?$'
)



class HelpBrowser(QWidget):
    def __init__(self):
        super(HelpBrowser, self).__init__()

        self.ui = Ui_HelpBrowser()
        self.ui.setupUi(self)

        if sys.platform.startswith('darwin'):
            self.setWindowFlags(Qt.Tool)

        self.ui.tabs.tabCloseRequested.connect(self.closeTab)

        self.openTabs = []

        self.shortcuts = [
            QShortcut(QKeySequence('Ctrl+W'), self, self.closeCurrentTab, self.closeCurrentTab),
            ]


        # home page
        # ---------

        self.openHelpTopic('/home') # welcome page.

        # make the home page un-closeable
        self.ui.tabs.tabBar().setTabButton(0, QTabBar.LeftSide, None)
        self.ui.tabs.tabBar().setTabButton(0, QTabBar.RightSide, None)



    @pyqtSlot()
    def closeCurrentTab(self):
        index = self.ui.tabs.currentIndex()
        if (index == 0):
            # close help browser
            self.hide()
            return
        
        self.closeTab(index)

    @pyqtSlot(int)
    def closeTab(self, index):
        if (index == 0):
            return
        del self.openTabs[index]
        self.ui.tabs.removeTab(index)

    @pyqtSlot()
    def openHelpTopicBySender(self):
        sender = self.sender()
        path = str(sender.property('helppath'))
        if (not path):
            logger.warning("Bad help topic path: %r", path)
            return

        self.openHelpTopic(path)
        

    @pyqtSlot('QUrl')
    def openHelpTopicUrl(self, url):
        urlstr = url.toString()
        if urlstr.startswith('#'):
            # pure fragment part, internal scroll
            self.sender().scrollToAnchor(urlstr[1:])
            return
        self.openHelpTopic(urlstr)


    @pyqtSlot(str)
    def openHelpTopic(self, url):

        logger.debug("Help: open topic at URL %s", url)

        # split off scheme and fragment, if any
        (scheme, netloc, path, query, fragment) = urlsplit(url)

        logger.debug("url parsed into (scheme,netloc,path,query,fragment) = %r",
                     (scheme, netloc, path, query, fragment))

        if scheme in ['http', 'https']:
            QDesktopServices.openUrl(QUrl(url))
            return
        
        if scheme and scheme != 'help':
            raise ValueError("Invalid URL scheme: {} [url={}]".format(scheme, url))

        # combine path with query string
        pathqs = path
        if query:
            pathqs += '?' + query

        def canonpath_check_fn(canonpath):
            urlcanon = 'help:'+canonpath
            logger.debug("requested page has canonical url %r", urlcanon)
            # check to see if the tab is already open
            for tab in self.openTabs:
                if (str(tab.property('helpurl')) == urlcanon):
                    # just raise this tab.
                    raise TabAlreadyOpen(tab)

        widget = None
        try:

            page = helppages.get_help_page(
                pathqs,
                html_table_width_px=TABLE_WIDTH,
                canonpath_check_fn=canonpath_check_fn, # raise tab if page already open
            )

            widget = HelpTopicPageWidget(page, self, self.ui.tabs)
            widget.setProperty("helpurl", 'help:'+page.canonpath())

            tabindex = self.ui.tabs.addTab(widget, widget.property('HelpTabTitle'))
            self.ui.tabs.setTabToolTip(tabindex, widget.property('HelpTabToolTip'))
            #self.ui.tabs.setCurrentIndex(tabindex)

            self.openTabs.append(widget)

        except TabAlreadyOpen as t:
            widget = t.widget

        if widget is None:
            logger.debug("Couldn't open help topic widget for %r", url)
            return

        self.ui.tabs.setCurrentWidget(widget)

        if fragment:
            # we were asked to scroll to a specific anchor
            widget.scrollToAnchor(fragment)






#
# initial help page for the GUI --- /home
#
def _get_help_page_home(pathitems, kwargs):

    if len(pathitems) != 0:
        raise ValueError("Invalid help path: {}".format('/'.join(kwargs['basepathitems']+pathitems)))

    def big_html_link(txt, link):
        return """<a class="biglink" href="{link}">{txt}</a>""".format(
            txt=txt, link=link
        )

    home_src = "<h1>Help &amp; Reference Browser</h1>\n\n"

    home_src += ("<p>" + big_html_link("Welcome to bibolamazi", "help:/general/welcome")
                 + " \N{EM DASH} " +
                 big_html_link("Annotated filter list", "help:/filters")
                 + " \N{EM DASH} " +
                 big_html_link("Bibolamazi online docs", "https://bibolamazi.readthedocs.org/")
                 + " \N{EM DASH} " +
                 big_html_link("Bibolamazi version information", "help:/general/version")
                 + " \N{EM DASH} " +
                 big_html_link("Bibolamazi command-line help", "help:/general/cmdline")
                 + "</p>\n\n")

    home_src += "<h2>Filters</h2>\n\n"

    from . import filterinstanceeditor # cyclic import if done @ top of module

    home_src += "<ul>\n"
    for filt in sorted(filterinstanceeditor.get_filter_list()):
        home_src += (
            "<li>&nbsp;<a href=\"help:/filter/{filtname}\"><b>{filtname}</b></a>\n".format(filtname=filt)
            )
    home_src += "</ul>\n"

    return helppages.HelpTopicPage.makeHtmlFragmentPage(home_src, title="Home", canonpath='/home')


#
# register help page dispatcher for /home
#
helppages.help_page_dispatchers['home'] = _get_help_page_home



# -*- coding: utf-8 -*-
################################################################################
#                                                                              #
#   This file is part of the Bibolamazi Project.                               #
#   Copyright (C) 2018 by Philippe Faist                                       #
#   philippe.faist@bluewin.ch                                                  #
#                                                                              #
#   Bibolamazi is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU General Public License as published by       #
#   the Free Software Foundation, either version 3 of the License, or          #
#   (at your option) any later version.                                        #
#                                                                              #
#   Bibolamazi is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#   GNU General Public License for more details.                               #
#                                                                              #
#   You should have received a copy of the GNU General Public License          #
#   along with Bibolamazi.  If not, see <http://www.gnu.org/licenses/>.        #
#                                                                              #
################################################################################

# Py2/Py3 support
from __future__ import unicode_literals, print_function
from past.builtins import basestring
from future.utils import python_2_unicode_compatible, iteritems
from builtins import range
from builtins import str as unicodestr

import re
import logging
from html import escape as htmlescape

import github

import bibolamazi.init

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

from .uiutils import ContextAttributeSetter, BlockedSignals
from .qtauto.ui_githubauthenticationdialog import Ui_GithubAuthenticationDialog

logger = logging.getLogger(__name__)


# ------------------------------------------------------------------------------


class GithubAuthenticationDialog(QDialog):
    def __init__(self, parent):
        super(GithubAuthenticationDialog, self).__init__(parent)

        self.ui = Ui_GithubAuthenticationDialog()
        self.ui.setupUi(self)

        # Button screenshots: take on hi dpi mac, scale image @ 80%, export to
        # png, note dimensions; specify 1/2 of the dimensions as image sizes
        # below so that pictures appear nice on retina displays

        self.ui.lbl.setText("""\
<html><head>
<style>
h1 { font-size: 120%; font-weight: bold; }
p { margin-bottom: 1em; }
ol, li { margin: 0px; padding: 0px; }
a { color: #4040ff; }
img { vertical-align: middle; }
</style>
<body>
<h1>Authenticate to Github</h1>

<p>Please visit <a
href="https://github.com/settings/tokens">https://github.com/settings/tokens</a>
and follow the following instructions to generate a personal access token.  This
procedure ensures that bibolamazi never sees your github password.</p>

<p>1. Click on the button <img
    src=":/pic/github_generate_new_personal_access_token.png" alt="Generate new
    token" width="128" height="29" style="vertical-align: middle; width: 128px; height: 29px;"></p>

<p>2. Give a name to the access token such as “bibolamazi access” and select the “repo” scope:</p>

<p><img src=":/pic/github_generate_new_personal_access_token_details.png" alt=""
        width="312" height="236" style="width: 312px; height: 236px;"></p>

<p>3. Scroll down and click <img src=":/pic/github_generate_new_personal_access_token_generatebtn.png"
    alt="Generate token" width="117" height="32"
    style="vertical-align: middle; width: 117px; height: 32px;"></p>

<p><b>4. Paste the access token in the field below:</b></p>

</body>
</html>""")
        
        self.ui.txtToken.textChanged.connect(self._update_gui_state)
        self._update_gui_state()

        self.adjustSize()
        self.resize(self.size() + QSize(150,80))


    def getAuthToken(self):
        return self.ui.txtToken.text().strip()

    @pyqtSlot()
    def _update_gui_state(self):

        with BlockedSignals(self.ui.txtToken):
            self.ui.txtToken.setText( self.ui.txtToken.text().strip() )

        self.ui.btns.button(QDialogButtonBox.Ok).setEnabled(
            self._is_valid_token(self.ui.txtToken.text())
        )
        

    def _is_valid_token(self, token):
        logger.debug("checking token = [...]%s", token[-4:])
        return (re.match(r'^[a-fA-F0-9]{32,}$', token if token else '') is not None)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'sourcelisteditor.ui'
#
# Created by: PyQt5 UI code generator 5.7.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_SourceListEditor(object):
    def setupUi(self, SourceListEditor):
        SourceListEditor.setObjectName("SourceListEditor")
        SourceListEditor.resize(382, 462)
        self.gridLayout = QtWidgets.QGridLayout(SourceListEditor)
        self.gridLayout.setVerticalSpacing(20)
        self.gridLayout.setObjectName("gridLayout")
        self.btnAddSource = QtWidgets.QPushButton(SourceListEditor)
        self.btnAddSource.setObjectName("btnAddSource")
        self.gridLayout.addWidget(self.btnAddSource, 2, 0, 1, 2)
        self.btnAddFavorite = QtWidgets.QPushButton(SourceListEditor)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/pic/bookmark.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.btnAddFavorite.setIcon(icon)
        self.btnAddFavorite.setObjectName("btnAddFavorite")
        self.gridLayout.addWidget(self.btnAddFavorite, 5, 0, 1, 2)
        self.lbl = QtWidgets.QLabel(SourceListEditor)
        self.lbl.setMinimumSize(QtCore.QSize(0, 100))
        self.lbl.setText("")
        self.lbl.setTextFormat(QtCore.Qt.RichText)
        self.lbl.setWordWrap(True)
        self.lbl.setObjectName("lbl")
        self.gridLayout.addWidget(self.lbl, 0, 0, 1, 2)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem, 3, 0, 1, 1)
        self.lblLinkInfo = QtWidgets.QLabel(SourceListEditor)
        self.lblLinkInfo.setAutoFillBackground(False)
        self.lblLinkInfo.setStyleSheet("QLabel { font-style: italic; background-color: rgba(0,0,0,10%); }")
        self.lblLinkInfo.setText("")
        self.lblLinkInfo.setWordWrap(True)
        self.lblLinkInfo.setObjectName("lblLinkInfo")
        self.gridLayout.addWidget(self.lblLinkInfo, 1, 0, 1, 2)

        self.retranslateUi(SourceListEditor)
        QtCore.QMetaObject.connectSlotsByName(SourceListEditor)

    def retranslateUi(self, SourceListEditor):
        _translate = QtCore.QCoreApplication.translate
        SourceListEditor.setWindowTitle(_translate("SourceListEditor", "Form"))
        self.btnAddSource.setText(_translate("SourceListEditor", "Add Independent Source"))
        self.btnAddFavorite.setToolTip(_translate("SourceListEditor", "Add this source list to your favorites"))
        self.btnAddFavorite.setText(_translate("SourceListEditor", "add souce && alternatives to favorites"))

from . import bibolamazi_res_rc



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'githubreposelector.ui'
#
# Created by: PyQt5 UI code generator 5.12.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_GithubRepoSelector(object):
    def setupUi(self, GithubRepoSelector):
        GithubRepoSelector.setObjectName("GithubRepoSelector")
        GithubRepoSelector.resize(430, 138)
        self.gridLayout = QtWidgets.QGridLayout(GithubRepoSelector)
        self.gridLayout.setObjectName("gridLayout")
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem, 2, 0, 1, 1)
        self.stk = QtWidgets.QStackedWidget(GithubRepoSelector)
        self.stk.setObjectName("stk")
        self.pageUsername = QtWidgets.QWidget()
        self.pageUsername.setObjectName("pageUsername")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.pageUsername)
        self.gridLayout_2.setContentsMargins(0, 0, 0, 0)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.txtUser = QtWidgets.QLineEdit(self.pageUsername)
        self.txtUser.setObjectName("txtUser")
        self.gridLayout_2.addWidget(self.txtUser, 1, 0, 1, 1)
        self.label = QtWidgets.QLabel(self.pageUsername)
        self.label.setObjectName("label")
        self.gridLayout_2.addWidget(self.label, 0, 0, 1, 1)
        self.stk.addWidget(self.pageUsername)
        self.pageRepo = QtWidgets.QWidget()
        self.pageRepo.setObjectName("pageRepo")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.pageRepo)
        self.gridLayout_3.setContentsMargins(0, 0, 0, 0)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.lblPromptRepo = QtWidgets.QLabel(self.pageRepo)
        self.lblPromptRepo.setObjectName("lblPromptRepo")
        self.gridLayout_3.addWidget(self.lblPromptRepo, 0, 0, 1, 1)
        self.cbxRepos = QtWidgets.QComboBox(self.pageRepo)
        self.cbxRepos.setEditable(True)
        self.cbxRepos.setObjectName("cbxRepos")
        self.gridLayout_3.addWidget(self.cbxRepos, 1, 0, 1, 1)
        self.stk.addWidget(self.pageRepo)
        self.gridLayout.addWidget(self.stk, 1, 0, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.btnCancel = QtWidgets.QPushButton(GithubRepoSelector)
        self.btnCancel.setObjectName("btnCancel")
        self.horizontalLayout.addWidget(self.btnCancel)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.btnBack = QtWidgets.QPushButton(GithubRepoSelector)
        self.btnBack.setObjectName("btnBack")
        self.horizontalLayout.addWidget(self.btnBack)
        self.btnNext = QtWidgets.QPushButton(GithubRepoSelector)
        self.btnNext.setObjectName("btnNext")
        self.horizontalLayout.addWidget(self.btnNext)
        self.btnOk = QtWidgets.QPushButton(GithubRepoSelector)
        self.btnOk.setObjectName("btnOk")
        self.horizontalLayout.addWidget(self.btnOk)
        self.gridLayout.addLayout(self.horizontalLayout, 3, 0, 1, 1)

        self.retranslateUi(GithubRepoSelector)
        self.stk.setCurrentIndex(1)
        self.btnCancel.clicked.connect(GithubRepoSelector.reject)
        QtCore.QMetaObject.connectSlotsByName(GithubRepoSelector)

    def retranslateUi(self, GithubRepoSelector):
        _translate = QtCore.QCoreApplication.translate
        GithubRepoSelector.setWindowTitle(_translate("GithubRepoSelector", "Dialog"))
        self.label.setText(_translate("GithubRepoSelector", "Github user name / organization:"))
        self.lblPromptRepo.setText(_translate("GithubRepoSelector", "Repositories for user %s"))
        self.btnCancel.setText(_translate("GithubRepoSelector", "cancel"))
        self.btnBack.setText(_translate("GithubRepoSelector", "◀ back"))
        self.btnNext.setText(_translate("GithubRepoSelector", "next ▶"))
        self.btnOk.setText(_translate("GithubRepoSelector", "OK"))


